/*
 * Copyright 2006, 2008 Hristo Koshutanski, Aliaksandr Lazouski, Associazione CREATE-NET, and Universita' degli Studi di Trento.  
 * 
 * This file is part of iAccess authorization service software (called iAccess).
 * iAccess is free software: you can redistribute it and/or modify 
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * iAccess is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with iAccess.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.iaccess.CredentialManager;

import org.bouncycastle.asn1.*;
import org.bouncycastle.asn1.x509.GeneralName;
import org.bouncycastle.x509.*;

/**
 * Provides a logical representation of a credential certificate. 
 * Defines a transformation to logic level credential tokens.
 */
public class CertificateLogical {
    private String holder;
    private String issuer;
    private String attribute;        
    
    /** Creates a new instance of CertificateLogical */
    public CertificateLogical() {
        this.holder = new String();
        this.issuer = new String();
        this.attribute = new String();
    }
    
    /** Creates a new instance of CertificateLogical by string representation of holder, attribute, and issuer. */
    public CertificateLogical( String holder, String issuer,   String attribute) {
        this.holder = holder;
        this.issuer = issuer;
        this.attribute = attribute;
    }
    
    
    public static String makeLogical(String cn)
    {
        cn = cn.trim().replaceAll("\\s+","_");
        cn = cn.replaceAll("\\-","_");
        cn = cn.replaceAll("\\\\","_");
        cn = cn.replaceAll("/","_");
        cn = cn.replaceAll("\\@","_");
        cn = cn.replaceAll("\\.","");
        cn = cn.replaceAll("\\,","");
        cn = cn.replaceAll("\\?","");
        cn = cn.replaceAll("\\!","");
        cn = cn.replaceAll("\\\"","");
        cn = cn.replaceAll("\\\'","");
        cn = cn.replaceAll("\\=","");
        cn = cn.replaceAll("\\%","");
        cn = cn.replaceAll("\\$","");
        return cn;
    }
    
    /** Creates a new instance of CertificateLogical from X509V2AttributeCertificate.*/
    public CertificateLogical( X509V2AttributeCertificate certificate) {
     try {
            DERIA5String attrString  = (DERIA5String) GeneralName.getInstance( certificate.getAttributes()[0].getValues()[0] ).getName();          
            this.holder = certificate.getHolder().getEntityNames()[0].getName().toLowerCase(); 
            if(this.holder.indexOf("cn=")>-1)
               this.holder = this.holder.substring(this.holder.indexOf("cn=")+3);
            else
            {
                this.holder=null;
                this.issuer=null;
                this.attribute=null;
                throw new Exception("Missing CN field of X500Holder.");
            }
            if(this.holder.indexOf("=") > -1)
            {
                this.holder = this.holder.substring(0, this.holder.indexOf("="));
                this.holder = this.holder.substring(0, this.holder.lastIndexOf(","));
                this.holder = makeLogical(this.holder);
            }
            else
                this.holder = makeLogical( this.holder ); 

            this.issuer = certificate.getIssuer().getPrincipals()[0].getName().toLowerCase();
            
            if(this.issuer.indexOf("cn=")>-1)
            { 
                this.issuer = this.issuer.substring(this.issuer.indexOf("cn=")+3);
                
                if(this.issuer.indexOf("=") > -1)
                {
                    this.issuer = this.issuer.substring(0, this.issuer.indexOf("="));
                    this.issuer = this.issuer.substring(0, this.issuer.lastIndexOf(","));
                    this.issuer = makeLogical(this.issuer);
                }
                else
                    this.issuer = makeLogical(this.issuer);
            }
            else
            {
                if(this.issuer.indexOf("o=")==-1)
                {
                    this.holder=null;
                    this.issuer=null;
                    this.attribute=null;
                    throw new Exception("Missing CN and O fields of X500Issuer.");
                }
                this.issuer = this.issuer.substring(this.issuer.indexOf("o=")+2);
                if(this.issuer.indexOf("=") > -1)
                {
                    this.issuer = this.issuer.substring(0, this.issuer.indexOf("="));
                    this.issuer = this.issuer.substring(0, this.issuer.lastIndexOf(","));
                    this.issuer = makeLogical(this.issuer);
                }
                else
                    this.issuer = makeLogical(this.issuer);
            }

            this.attribute = makeLogical(attrString.getString().toLowerCase());

       } catch (Exception e) { 
            System.err.println("IACCESS: ERROR: CertificateLogical.CertificateLogical(X509V2AttrCert): " + e);
        }
    }    
    
    
    /** Creates a new instance of CertificateLogical from a java.security.cert.X509Certificate. */     
    public CertificateLogical(java.security.cert.X509Certificate certificate) {
        try{
            this.holder = certificate.getSubjectX500Principal().getName().trim().toLowerCase();
            
            if(this.holder.indexOf("cn=")>-1)
               this.holder = this.holder.substring(this.holder.indexOf("cn=")+3);
            else
            {
                this.holder=null;
                this.issuer=null;
                this.attribute=null;
                throw new Exception("Missing CN field of X500Holder.");
            }

            if(this.holder.indexOf("=") > -1)
            {
                this.holder = this.holder.substring(0, this.holder.indexOf("="));
                this.holder = this.holder.substring(0, this.holder.lastIndexOf(","));
                this.holder = makeLogical(this.holder);
            }
            else
                this.holder = makeLogical( this.holder ); 
            
            this.issuer =  certificate.getIssuerX500Principal().getName().trim().toLowerCase();
            
            if(this.issuer.indexOf("cn=")>-1)
            { 
                this.issuer = this.issuer.substring(this.issuer.indexOf("cn=")+3);
                
                if(this.issuer.indexOf("=") > -1)
                {
                    this.issuer = this.issuer.substring(0, this.issuer.indexOf("="));
                    this.issuer = this.issuer.substring(0, this.issuer.lastIndexOf(","));
                    this.issuer = makeLogical(this.issuer);
                }
                else
                    this.issuer = makeLogical(this.issuer);
            }
            else
            {
                if(this.issuer.indexOf("o=")==-1)
                {
                    this.holder=null;
                    this.issuer=null;
                    this.attribute=null;
                    throw new Exception("Missing CN and O fields of X500Issuer.");
                }
                this.issuer = this.issuer.substring(this.issuer.indexOf("o=")+2);
                if(this.issuer.indexOf("=") > -1)
                {
                    this.issuer = this.issuer.substring(0, this.issuer.indexOf("="));
                    this.issuer = this.issuer.substring(0, this.issuer.lastIndexOf(","));
                    this.issuer = makeLogical(this.issuer);
                }
                else
                    this.issuer = makeLogical(this.issuer);
            }
            
            this.attribute = null;
        }
        catch(Exception e)
        {
            System.err.println("IACCESS: ERROR: CertificateLogical.CertificateLogical(X509Certificate): " + e);
        }
    } 
                
    public String getHolder() {  return holder;  }
    public void setHolder(String holder) { this.holder = holder; }
    
    public String getIssuer() {  return issuer; }
    public void setIssuer(String issuer) { this.issuer = issuer; }
    
    public String getAttribute() { return attribute; }
    public void setAttribute(String attribute) { this.attribute = attribute; }
    
}
