/*
 * Copyright 2006, 2008 Hristo Koshutanski, Aliaksandr Lazouski, Associazione CREATE-NET, and Universita' degli Studi di Trento.  
 * 
 * This file is part of iAccess authorization service software (called iAccess).
 * iAccess is free software: you can redistribute it and/or modify 
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * iAccess is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with iAccess.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.iaccess.CredentialManager;

import org.iaccess.TNProtocol.TSessionInfo;

import java.util.*;
import java.io.*;

/**
 * Keeps information of users/opponents active credentials.
 * Provides necessary methods for retrieving profile information.
 * @see org.iaccess.TNProtocol.TSessionInfo
 */
public class UserProfile {
    
    static String ProfilesDB = CredentialsConfig.ProfilesDB;
    static File profileDir = new File(  ProfilesDB );
    private static Hashtable userProfiles = new Hashtable();
  
  
    public UserProfile() {
        
    }

    /**
     * Updates a user profile given a user identity and a user credential.
     * If there is no profile for the given identity , the function creates one.
     * The credential is added if it does not exist in the profile.
     */
    public static synchronized void updateOngoingProfile(String identity, String credential) {
        try {
            if( TSessionInfo.CredAnyValueTerm != null && 
                identity.compareTo(TSessionInfo.CredAnyValueTerm)==0)
                return;
            if ( ! userProfiles.containsKey(identity) )
                userProfiles.put(identity , new ArrayList() );
            if ( ! ((ArrayList) userProfiles.get(identity)).contains(credential) ) {
                ((ArrayList) userProfiles.get(identity)).add(credential);
            System.out.println("IACCESS: Updated user profile: " + identity + " with: " + credential + "");}
        } catch (Exception e) { System.err.println("IACCESS: EXCEPTION: UserProfile.updateOngoingProfile:" + e.toString());}
    }
    

    /**
     * Checks if a user credential exists in the user profile.
     * @param identity a user identity
     * @param credential a user credential of a negotiaiton level format (id,issuer or id,attribute,issuer)
     * @return if a credential exists in a user profile and is valid for a given session time validity.
     */
    public static synchronized boolean containts(String identity, String credential) {
        boolean result = false;
        try {
            result = ((ArrayList) userProfiles.get(identity)).contains(credential);
        } catch (Exception e) {  
            System.err.println("IACCESS: EXCEPTION: UserProfile.containts: " + e.toString());
        }        
        
        return result;
    }

    /**
     * Checks if a user credential exists in the user profile, and if the user credential has validity
     * longer than a current session expiration time.
     * @param credential a user credential of a negotiaiton level format (id,issuer or id,attribute,issuer)
     * @param sessionValidNotAfter a current session espiration time
     * @return if a credential exists in a user profile and is valid for a given session time validity.
     */
    public static synchronized boolean containts(String credential, long sessionValidNotAfter) {
        boolean result = false;
        try {
            String identity = credential.substring(0,credential.indexOf(','));
            result = ((ArrayList) userProfiles.get(identity)).contains(credential);
            if(result && CertificateManager.CredentialsValidity.containsKey(credential) &&
               sessionValidNotAfter < Long.parseLong((String)CertificateManager.CredentialsValidity.get(credential)) )
                result = true;
            else
                result = false;
                
        } catch (Exception e) {  
            System.err.println("IACCESS: EXCEPTION: UserProfile.containts: " + e.toString());
        }        
        
        return result;
    }
    
    
    
    /**
     * Returns an identity credential (logic format) of an SSL identity token from a given user identity.
     * Returns null if no SSL token is found. <p>On each SSL connection establishment by an opponent,
     * its SSL logic identity token is stored.
     * <p>If "john_couk" identity is given, the function would return "john_couk,california_state_ca".
     * <p>We assume the use uses a single identity token (for SSL) during a negotiation process.
     * @param identity logic level identity
     * @return a logic format identity token of the used X.509 token for SSL connection.
     */
    public static synchronized String getEntityLogicIDCredOfSSL(String identity)
    {
        String cred = null;
        boolean fl = false;
        try{
                Enumeration allEntities =  userProfiles.keys();
                for (; allEntities.hasMoreElements(); ) 
                {
                    String currentEntity = (String)allEntities.nextElement(); 
                    if(currentEntity.indexOf("CN=")>-1)
                    {
                        Object[] currentEntityCreds = ((ArrayList)userProfiles.get(currentEntity)).toArray();
                        for(int i=0;i<currentEntityCreds.length;i++)
                            if(((String)currentEntityCreds[i]).startsWith(identity))
                            {
                                cred = new String((String)currentEntityCreds[i]);
                                fl = true;
                            }
                    }
                    if(fl) break;
                }
        }
        catch(Exception e)
        {
            System.err.println("IACCESS: EXCEPTION: CredentialManager.UserProfile.getEntityLogicIDCredOfSSL(): " +e);
        }
        
        return cred;
    }

    /**
     * Returns the current set of active/presented credentials of a given identity. Logic level format.
     */
    public static synchronized Vector getCredentials(String identity) {
        ArrayList foundCreds = null;
        Vector toReturn=new Vector();
        try {
            foundCreds = (ArrayList) userProfiles.get(identity);
            if(foundCreds!=null)
              for(int i=0;i<foundCreds.size();i++)
                  toReturn.addElement((String)foundCreds.get(i));
        } 
        catch (Exception e) 
        { 
            System.err.println("IACCESS: EXCEPTION: UserProfile.getCredentials: "+e);
        }         
        return toReturn;
    }


    /**
     * Returns a set of presented credentials of a given identity (logic level format) but valid for a current
     * session time validity.
     */
    public static synchronized Vector getCredentials(String identity, long sessionValidNotAfter) {
        ArrayList foundCreds = null;
        Vector toReturn = new Vector();
        try {
            foundCreds = (ArrayList) userProfiles.get(identity);
            if(foundCreds!=null)
              for(int i=0;i<foundCreds.size();i++)
              {
                  Object certVal = CertificateManager.CredentialsValidity.get(foundCreds.get(i));
                  if(certVal!=null)
                  {
                     long certValidity = Long.parseLong((String)certVal);
                     if(sessionValidNotAfter < certValidity)
                        toReturn.addElement((String)foundCreds.get(i));
                  }
              }
        } 
        catch (Exception e) 
        { 
            System.err.println("IACCESS: EXCEPTION: UserProfile.getCredentials: "+e);
        }         
        return toReturn;
    }
    
    /**
     * Looks up in the hashtable to find the X500 Name of the given logic identity.
     * The X500 name is stored from the SSL identity token used by the opponent for establishing SSL connections.
     */
    public static synchronized String getX500Name(String Identity)
    {
        try
        {
                Enumeration allEntities =  userProfiles.keys();
                for (; allEntities.hasMoreElements(); ) 
                {
                    String nextEntity = (String)allEntities.nextElement(); 
                    Object[] nextEntityCreds = ((ArrayList)userProfiles.get(nextEntity)).toArray();
                    for(int i=0;i<nextEntityCreds.length;i++)
                    {
                        if( (nextEntity.indexOf("CN=")>-1) && ((String)nextEntityCreds[i]).startsWith(Identity) 
                           )
                        {
                            return new String(nextEntity);
                        }
                    }
                }
        }
        catch(Exception e)
        {
            System.err.println("IACCESS: EXCEPTION: CredentialManager.UserProfile getX500Name: " + e.toString());
        }
        
         return null;
    }
}
