/*
 * Copyright 2006, 2008 Hristo Koshutanski, Aliaksandr Lazouski, Associazione CREATE-NET, and Universita' degli Studi di Trento.  
 * 
 * This file is part of iAccess authorization service software (called iAccess).
 * iAccess is free software: you can redistribute it and/or modify 
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * iAccess is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with iAccess.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.iaccess.LogicalLayer;

import java.util.regex.*;
import java.util.Date;
import javax.xml.parsers.DocumentBuilderFactory;
import org.w3c.dom.*;
import java.io.*;
import java.util.Vector;
import java.util.LinkedList;
import org.xml.sax.InputSource;

import org.iaccess.TNProtocol.TRequest;

/**
 * Encapsulates the logic level computations. It functions as a wrapper to the DLV solver used
 * as an underlying engine for deduction and abduction reasonings. The configuration file is config_dlv.xml. All computations on the dlv system are
 * taken from the configuration file. Instead of recompilling the file, one can change some of the imput parameters to DLV in the
 * configuration file. For exmaple, from 'cautious' to 'brave' reasoning, as explained in the doDeduction() function.
 * <p> The class implements the interactive access control algorithm, function computeAccessDecision(),
 * and the stepwise disclosure of missing credentials, function doStepwiseDisclosure().
 * <p> Improvements on the interactive access control, and, on the complementing it, stepwise disclosure
 * have been done with respect to the published negotiation scheme in the JNSM paper.
 * <p> The improvements mainly concern the way disclosure policy is treated in order to avoid
 * endless loops. The last may occur if a client declines to present a missing credential that is not part of a solution for a resource, but is part
 * of a fine-grained disclosure control (stepwise computing).
 * For such cases, in the SetUp function we prepare the disclosure policy in a way of not chaining any credential in a head of a rule
 * that belongs to the set of declined credentials. Now, during a stepwise negotiation/computing whenever a client declines to present credentials
 * the access decision will take it into account, i.e., any access decision for a missing set of credentials is computed as all credentials of the missing set are
 * properly chained (by the disclosure policy) without any declined credentials in a chain.
 * <p>Some more details are given for each of the functions in the class.
 * <p> ResultOfOperation class is used to interface the results of logic level computations and the negotiation level.
 * @see org.iaccess.TNProtocol.iAccessNegotiation
 * @see org.iaccess.LogicalLayer.ResultOfOperation
 */
public class BasicAccessControlAlgorithm {

       public static volatile boolean IsInitialized = false;

       private boolean IsRequestForCertificate = false;
       private String FileStr;

       private Vector Cert_Presented, Cert_Declined, 
               Cred_Declined, Cred_Presented, 
               Cred_Presented_With_AnyValueTerm,
               Cert_Presented_With_AnyValueTerm;
       private TRequest Request;
       private File deductionPositiveFactsFile,
       deductionQueryFile,
       abductionHypothesesFile,
       abductionObservationsFile,
       stepwiseDisclosurePolicy,
       stepwiseDisclosurePolicyWithComplementRules,
       stepwiseDisclosurePolicyForOneStep, //Only complemented rules it is not a policy
       queryOneStep,
       oneStepHypotheses,
       oneStepObservations,
       deductionPositiveFactsFileWithAnyValueTermCreds,//this is used only in doAbduction(). 
       requestAsFactFile; //this is used for evaluation of the credential policy

   /** If true, the algorithm selects minimal sets with first weight-minimality then set-cardinality,
    * else first minimal set-cardinality and then weight-minimality.
    * The field is static and shared among all objects of this class. It is initialized once.
    */
   private static boolean CriterionRoleMinimalitySetCardinality; 
   
   private static String
                //String patterns for a common use
                PatternContentStr        = "\\$\\{content\\}",
                PatternLeftBraceStr      = "\\(",
                PatternRightBraceStr     = "\\)",
                PatternStr2LeftBraceStr  = "\\\\(",
                PatternStr2RightBraceStr = "\\\\)",
                PatternWhiteSpacesStr    = "\\s+",

                // need them for abduction process
                hypothesesFileExt   = ".hyp",
                observationsFileExt = ".obs",

                // Initializing them from config.xml file. They have the same name as the nodes in that xml file.
                CredentialsDelimiter = "\\,",
                OSFileDelimiter = "\\",
                Prefix4DeclinedCred = "declined", //internal use only.
                PredicateName4SystemTimeNow,
                PredicateName4TNTimeByNow,
                Rule4NotChainCred,
                Rule4NotChainCert,
                WorkingDirectory,
                DLVDirectory,
                DLVExecutableName,
                DLVDeductionCommandPromptOptions,
                DLVAbductionCommandPromptOptions,
                DLVComputeAbduciblesCommandPromptOptions,
                DLVOneStepDeductionCommandPromptOptions,
                DeductionServicePolicyFiles,
                DeductionCredentialPolicyFiles,
                ComputeAbduciblesPolicyFiles,
                AbductionPolicyFiles,
                AbductionCredentialPolicyFiles,
                OneStepAbductionPolicyFiles,
                DLVKeywords4SUccessfulDeduction,
                DLVTemplate4Credential,
                DLVExtendedTemplate4Credential,
                DLVTemplate4Certificate,
                DLVTemplate4Service,
                DLVOneStepPrefix4Credentials,
                DLVAbductionOneStepCommandPromptOptions,

                // Configuration file path.
                configFilePath;

                // Commonly used patterns
        private static Pattern PatternLeftBrace,
                    PatternRightBrace,
                    PatternContent,
                    PatternWhiteSpaces;

        public BasicAccessControlAlgorithm(){}


        /**
         * Runs only once for all negotiations. It parses the configuration file and prepares some rules used in the
         * stepwise computing. This function is complemented by the SetUp function which is run before any decision to be taken.
         */
        public static synchronized void LoadConfiguration(String configPath, boolean MinimalityCriterion, String ConfigRootDir, String RootDir, int os) throws Exception{
                if (IsInitialized)
                        return;

        configFilePath = configPath;
        CriterionRoleMinimalitySetCardinality = MinimalityCriterion;

        //Initializing Patterns
        PatternLeftBrace   = Pattern.compile(PatternLeftBraceStr);
        PatternRightBrace  = Pattern.compile(PatternRightBraceStr);
        PatternContent     = Pattern.compile(PatternContentStr);
        PatternWhiteSpaces = Pattern.compile(PatternWhiteSpacesStr);


        File configFile = new File(configFilePath);
        if(!configFile.exists())
            throw new Exception("The specified DLVConfigFile does not exist!");
        try{
                BufferedReader configFileToModify = new BufferedReader( new FileReader(configFilePath) );
                String readLine,
                       newConfigFile = "";
                while((readLine = configFileToModify.readLine()) != null)
                    newConfigFile += readLine;
                //If Windows change file separator to "\" , otherwise change to "/"
                String myrootDir = new String(RootDir);
                if(os==2)
                {
                    //Because root replacement part below removes C:\file\ to C:file. 
                    myrootDir = myrootDir.replace('\\',';');   
                    newConfigFile = Pattern.compile(ConfigRootDir).matcher(newConfigFile).replaceAll(myrootDir);
                    newConfigFile = newConfigFile.replace(';','\\');
                }
                else
                {
                    newConfigFile = Pattern.compile(ConfigRootDir).matcher(newConfigFile).replaceAll(myrootDir);
                    newConfigFile = newConfigFile.replace(';','/');
                }
                //Instantiating the root dir variable in the config file.
                newConfigFile = Pattern.compile(ConfigRootDir).matcher(newConfigFile).replaceAll(myrootDir);
                // Create a builder factory
                DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		factory.setValidating(false);
		// Create the builder and parse the confogiration from StramingInput
		Document doc = factory.newDocumentBuilder().parse(new InputSource( new StringReader(newConfigFile)));

                // Initializing ....
                WorkingDirectory                         = doc.getDocumentElement().getElementsByTagName("WorkingDirectory").item(0).getFirstChild().getNodeValue().trim();
                DLVDirectory                             = doc.getDocumentElement().getElementsByTagName("DLVDirectory").item(0).getFirstChild().getNodeValue().trim();
                DLVExecutableName                        = doc.getDocumentElement().getElementsByTagName("DLVExecutableName").item(0).getFirstChild().getNodeValue().trim();
                DLVDeductionCommandPromptOptions         = doc.getDocumentElement().getElementsByTagName("DLVDeductionCommandPromptOptions").item(0).getFirstChild().getNodeValue().trim();
                DLVAbductionCommandPromptOptions         = doc.getDocumentElement().getElementsByTagName("DLVAbductionCommandPromptOptions").item(0).getFirstChild().getNodeValue().trim();
                DLVComputeAbduciblesCommandPromptOptions = doc.getDocumentElement().getElementsByTagName("DLVComputeAbduciblesCommandPromptOptions").item(0).getFirstChild().getNodeValue().trim();
                DeductionServicePolicyFiles              = doc.getDocumentElement().getElementsByTagName("DeductionServicePolicyFiles").item(0).getFirstChild().getNodeValue().trim();
                DeductionCredentialPolicyFiles           = doc.getDocumentElement().getElementsByTagName("DeductionCredentialPolicyFiles").item(0).getFirstChild().getNodeValue().trim();
                ComputeAbduciblesPolicyFiles             = doc.getDocumentElement().getElementsByTagName("ComputeAbduciblesPolicyFiles").item(0).getFirstChild().getNodeValue().trim();
                AbductionPolicyFiles                     = doc.getDocumentElement().getElementsByTagName("AbductionPolicyFiles").item(0).getFirstChild().getNodeValue().trim();
                AbductionCredentialPolicyFiles           = doc.getDocumentElement().getElementsByTagName("AbductionCredentialPolicyFiles").item(0).getFirstChild().getNodeValue().trim();
                OneStepAbductionPolicyFiles              = doc.getDocumentElement().getElementsByTagName("OneStepAbductionPolicyFiles").item(0).getFirstChild().getNodeValue().trim();
                DLVKeywords4SUccessfulDeduction          = doc.getDocumentElement().getElementsByTagName("DLVKeywords4SUccessfulDeduction").item(0).getFirstChild().getNodeValue().trim();
                DLVTemplate4Credential                   = doc.getDocumentElement().getElementsByTagName("DLVTemplate4Credential").item(0).getFirstChild().getNodeValue().trim();
                DLVExtendedTemplate4Credential           = doc.getDocumentElement().getElementsByTagName("DLVExtendedTemplate4Credential").item(0).getFirstChild().getNodeValue().trim();
                DLVTemplate4Certificate                  = doc.getDocumentElement().getElementsByTagName("DLVTemplate4Certificate").item(0).getFirstChild().getNodeValue().trim();
                DLVTemplate4Service                      = doc.getDocumentElement().getElementsByTagName("DLVTemplate4Service").item(0).getFirstChild().getNodeValue().trim();

                DLVAbductionOneStepCommandPromptOptions  = doc.getDocumentElement().getElementsByTagName("DLVAbductionOneStepCommandPromptOptions").item(0).getFirstChild().getNodeValue().trim();
                DLVOneStepDeductionCommandPromptOptions  = doc.getDocumentElement().getElementsByTagName("DLVOneStepDeductionCommandPromptOptions").item(0).getFirstChild().getNodeValue().trim();
                DLVOneStepPrefix4Credentials             = doc.getDocumentElement().getElementsByTagName("DLVOneStepPrefix4Credentials").item(0).getFirstChild().getNodeValue().trim();
                PredicateName4SystemTimeNow              = doc.getDocumentElement().getElementsByTagName("PredicateName4SystemTimeNow").item(0).getFirstChild().getNodeValue().trim();
                PredicateName4TNTimeByNow                = doc.getDocumentElement().getElementsByTagName("PredicateName4TNTimeByNow").item(0).getFirstChild().getNodeValue().trim();
        }
        catch(Exception e)
        {
            throw new Exception("Parsing the DLV config file exception: " + e);
        }

        
            String replaceStr = " ";

            // Re-Initializing some entries from config.file
            // replacing white dublicate white spaces with one space, including new line etc.
            AbductionPolicyFiles           = PatternWhiteSpaces.matcher(AbductionPolicyFiles).replaceAll(replaceStr);
            ComputeAbduciblesPolicyFiles   = PatternWhiteSpaces.matcher(ComputeAbduciblesPolicyFiles).replaceAll(replaceStr);
            DeductionServicePolicyFiles    = PatternWhiteSpaces.matcher(DeductionServicePolicyFiles).replaceAll(replaceStr);
            DeductionCredentialPolicyFiles = PatternWhiteSpaces.matcher(DeductionCredentialPolicyFiles).replaceAll(replaceStr);
            AbductionCredentialPolicyFiles = PatternWhiteSpaces.matcher(AbductionCredentialPolicyFiles).replaceAll(replaceStr);
            OneStepAbductionPolicyFiles    = PatternWhiteSpaces.matcher(OneStepAbductionPolicyFiles).replaceAll(replaceStr);
        
        switch(os)
        {   // 1 for Linux
            case 1:
                DLVDirectory += "/linux";
                DLVExecutableName += ".bin";
                OSFileDelimiter = "/";
                break;
            // 2 for Windows    
            case 2:
                DLVDirectory += "\\windows";
                DLVExecutableName += ".exe";
                OSFileDelimiter = "\\";
                break;
            // 3 for MacOSX   
            case 3:
                DLVDirectory += "/macosx";
                DLVExecutableName += ".bin";
                OSFileDelimiter = "/";
                break;
            default:
                throw new Exception("Unsupported Operating System Type.");
        }
        // Parsing input declarations & credentials ....
        // Checking the important directories ...
        File workDir = new File(WorkingDirectory);
        if(workDir.isDirectory())
            WorkingDirectory = workDir.getPath();
        else
            throw new Exception("The specified Working Directory does not exist!");
        File DLVDir = new File (DLVDirectory);
        if(DLVDir.isDirectory())
            DLVDirectory = DLVDir.getPath();
        else
            throw new Exception("The specified DLV Directory does not exist!");


        String tmpCred = PatternContent.matcher(DLVTemplate4Credential).replaceAll("X,Y,Z");
        String tmpCert = PatternContent.matcher(DLVTemplate4Certificate).replaceAll("X,Y");
        Rule4NotChainCred = tmpCred + ":-" + DLVOneStepPrefix4Credentials + tmpCred + ", not " + Prefix4DeclinedCred + DLVOneStepPrefix4Credentials + tmpCred + ".";
        Rule4NotChainCert = tmpCert + ":-" + DLVOneStepPrefix4Credentials + tmpCert + ", not " + Prefix4DeclinedCred + DLVOneStepPrefix4Credentials + tmpCert + ".";

        
        IsInitialized = true;

        }

        /**
         * Deletes temporal files generated to support internal computations.
         */
    private void deleteTempFiles()
        {
             try
             {
                 if(deductionPositiveFactsFile != null && deductionPositiveFactsFile.exists())
                    deductionPositiveFactsFile.delete();

                 if(deductionQueryFile != null && deductionQueryFile.exists())
                    deductionQueryFile.delete();

                 if(abductionHypothesesFile != null && abductionHypothesesFile.exists())
                    abductionHypothesesFile.delete();

                 if(abductionObservationsFile != null && abductionObservationsFile.exists())
                    abductionObservationsFile.delete();


                 if(stepwiseDisclosurePolicy != null && stepwiseDisclosurePolicy.exists())
                    stepwiseDisclosurePolicy.delete();

                 if(stepwiseDisclosurePolicyWithComplementRules != null && stepwiseDisclosurePolicyWithComplementRules.exists())
                    stepwiseDisclosurePolicyWithComplementRules.delete();

                 if(stepwiseDisclosurePolicyForOneStep != null && stepwiseDisclosurePolicyForOneStep.exists())
                    stepwiseDisclosurePolicyForOneStep.delete();

                 if(queryOneStep != null && queryOneStep.exists())
                    queryOneStep.delete();

                 if(oneStepHypotheses != null && oneStepHypotheses.exists())
                    oneStepHypotheses.delete();

                 if(oneStepObservations != null && oneStepObservations.exists())
                    oneStepObservations.delete();
                 
                 if(deductionPositiveFactsFileWithAnyValueTermCreds != null && deductionPositiveFactsFileWithAnyValueTermCreds.exists())
                     deductionPositiveFactsFileWithAnyValueTermCreds.delete();
                 
                 if(requestAsFactFile != null && requestAsFactFile.exists())
                     requestAsFactFile.delete();
                 
             }
             catch(Exception e)
             {
                 System.out.println("IACCESS: EXCEPTION: While deleting temp files from the working directory:  " + e.toString());
                 // could be written in a log file as following warning: [date]: failed to delete temporary files!
             }
        }

        /**
         * A supporting function for string matching and replacement.
         */
        private String replacePatStr2Str(String patternStr, String inputStr, String replaceStr)
        {
             Pattern pattern = Pattern.compile(patternStr);
             Matcher matcher = pattern.matcher(inputStr);
             return matcher.replaceAll(replaceStr);
        }


        /**
         * Use this function before to run computeAccessDecision() and doStewiseDisclosure() functions.
         * This functions prepares the disclosure policy and the set of presented and declined credentials for
         * appropriate computations. We use credential's and certificate's templates to wrap around the values of credentials
         * from negotiation layer to the policy layer, transparantly to the negotiation.
         */
        public ResultOfOperation SetUp(TRequest Request, String ServiceRequest, Vector CP, Vector CN, String AnyValueTerm, long negStartTime)
        {
                ResultOfOperation resSetUp = new ResultOfOperation();

                try {
//                       Separate Credentials/Certificates and Replase White Spaces
                        Cert_Presented = new Vector(); Cred_Presented = new Vector();
                        Cert_Declined = new Vector();  Cred_Declined = new Vector();
//                        AllCredentialsInHeadOfRules = new Vector();
                        Cred_Presented_With_AnyValueTerm = new Vector();
                        Cert_Presented_With_AnyValueTerm = new Vector();

                        String InternalDelimiter = ",";
                        String replaceStr = "";
                        int FirstIndex = -1;

                        if (CP != null) {
                            for (int i = 0; i < CP.size(); i++){
                                        FirstIndex = CP.get(i).toString().indexOf(InternalDelimiter);
                                        if (FirstIndex == -1)
                                                throw new Exception("Could not identify the credential!");
                                        if (FirstIndex == CP.get(i).toString().lastIndexOf(InternalDelimiter))
                                        {   // If certificate by text,text
                                            if(CP.get(i).toString().indexOf(AnyValueTerm)==-1)
                                            {
                                                Cert_Presented.addElement(PatternWhiteSpaces.matcher(CP.get(i).toString()).replaceAll(replaceStr));
                                                String tmp1 = PatternWhiteSpaces.matcher(CP.get(i).toString()).replaceAll(replaceStr);
                                                tmp1 = AnyValueTerm + tmp1.substring(tmp1.indexOf(','));
                                                Cert_Presented_With_AnyValueTerm.addElement(tmp1);
                                            }
                                        }
                                        else
                                        {   // when a credential by text,text,text
                                            if(CP.get(i).toString().indexOf(AnyValueTerm)==-1)
                                            {
                                                Cred_Presented.addElement(PatternWhiteSpaces.matcher(CP.get(i).toString()).replaceAll(replaceStr));
                                                String tmp1 = PatternWhiteSpaces.matcher(CP.get(i).toString()).replaceAll(replaceStr);
                                                tmp1 = AnyValueTerm + tmp1.substring(tmp1.indexOf(','));
                                                Cred_Presented_With_AnyValueTerm.addElement(tmp1);
                                            }
                                        }
                               }
                        }

                        if (CN != null) {
                                FirstIndex = -1;
                                for (int i = 0; i < CN.size(); i++){
                                        FirstIndex = CN.get(i).toString().indexOf(InternalDelimiter);
                                        if (FirstIndex == -1)
                                                throw new Exception("Could not identify the credential!");
                                        if (FirstIndex == CN.get(i).toString().lastIndexOf(InternalDelimiter))
                                                Cert_Declined.addElement(PatternWhiteSpaces.matcher(CN.get(i).toString()).replaceAll(replaceStr));
                                        else
                                                Cred_Declined.addElement(PatternWhiteSpaces.matcher(CN.get(i).toString()).replaceAll(replaceStr));
                               }
                        }

//                       Analyze Request
                        if ( (Request.SessionID == null) && (Request.ServiceRequest == null) )
                            throw new Exception("Service is not Specified!");
                        else {
                                this.Request = Request;
                                this.Request.SessionID      = PatternWhiteSpaces.matcher(this.Request.SessionID).replaceAll(replaceStr);
                                this.Request.ServiceRequest = PatternWhiteSpaces.matcher(this.Request.ServiceRequest).replaceAll(replaceStr);
                        }

            FileStr = new String(this.Request.SessionID + ((org.iaccess.TNProtocol.iAccessNegotiation) Thread.currentThread()).getCounter());
            deductionPositiveFactsFile = new File(WorkingDirectory  + OSFileDelimiter + "_deductionPositiveFacts_" + FileStr);
            deductionQueryFile         = new File(WorkingDirectory  + OSFileDelimiter + "_deductionQuery_"         + FileStr);
            deductionPositiveFactsFileWithAnyValueTermCreds = new File(WorkingDirectory  + OSFileDelimiter + "_deductionPositiveFactsFileWithAnyValueTermCreds_" + FileStr);
            
            // Presenting it as a Stream in order to write bytes in it
            FileOutputStream deductionPositiveFactsStream = new FileOutputStream(deductionPositiveFactsFile);
            FileOutputStream deductionQueryStream = new FileOutputStream(deductionQueryFile);
            FileOutputStream deductionPositiveFactsFileWithAnyValueTermCredsStream = new FileOutputStream(deductionPositiveFactsFileWithAnyValueTermCreds);

            // Writing decl, cred, query ...
            // Writing declarations in a form suitable for the FORMAL MODEL in a FILE
            String TempStr;
            
            //Writing the CURRENT TIME Predicate
            String predicateTimeNow = (new Date()).toString().toLowerCase();
            predicateTimeNow = predicateTimeNow.replaceAll("\\s+",",");
            predicateTimeNow = predicateTimeNow.replace(':', ',');
            predicateTimeNow = PredicateName4SystemTimeNow + "("+predicateTimeNow+").";
            //Current Time Predicate: predicateTimeNow
            deductionPositiveFactsStream.write( predicateTimeNow.getBytes() );
            deductionPositiveFactsFileWithAnyValueTermCredsStream.write( predicateTimeNow.getBytes() );
            
            //Writing Current NegTime predicate
            String predicateTNTimeByNow = String.valueOf(System.currentTimeMillis() - negStartTime);
            predicateTNTimeByNow = PredicateName4TNTimeByNow + "("+predicateTNTimeByNow+").";
            //Current TN Time Predicate: predicateTNTimeByNow
            deductionPositiveFactsStream.write( predicateTNTimeByNow.getBytes() );
            deductionPositiveFactsFileWithAnyValueTermCredsStream.write( predicateTNTimeByNow.getBytes() );
            
            // we need to do this transformation only once, that is why I do it here.
            for(int i=0; i < Cert_Presented.size(); i++) {
                TempStr = PatternContent.matcher(DLVTemplate4Certificate).replaceAll(Cert_Presented.get(i).toString());
                Cert_Presented.setElementAt(TempStr,i);
                deductionPositiveFactsStream.write( (TempStr + ".").getBytes() );
                deductionPositiveFactsFileWithAnyValueTermCredsStream.write( (TempStr + ".").getBytes() );
            }

            // No need to do the crednetial teplate transformation here, and later it is done in different ways.
            // Store the transformation in a temp file
            for(int i=0; i < Cred_Presented.size(); i++) {
                TempStr = PatternContent.matcher(DLVTemplate4Credential).replaceAll(Cred_Presented.get(i).toString());
                deductionPositiveFactsStream.write( (TempStr + ".").getBytes() );
                deductionPositiveFactsFileWithAnyValueTermCredsStream.write( (TempStr + ".").getBytes() );
            }

            //Adding in a temp file also the presented credentials/certificates with anyvalue terms
            for(int i=0; i < Cred_Presented_With_AnyValueTerm.size(); i++) {
                TempStr = PatternContent.matcher(DLVTemplate4Credential).replaceAll(Cred_Presented_With_AnyValueTerm.get(i).toString());
                deductionPositiveFactsFileWithAnyValueTermCredsStream.write( (TempStr + ".").getBytes() );
            }
            for(int i=0; i < Cert_Presented_With_AnyValueTerm.size(); i++) {
                TempStr = PatternContent.matcher(DLVTemplate4Certificate).replaceAll(Cert_Presented_With_AnyValueTerm.get(i).toString());
                deductionPositiveFactsFileWithAnyValueTermCredsStream.write( (TempStr + ".").getBytes() );
            }
            
            
                // We need to do this transformation only once, that is why I do it here.
            for(int i=0; i < Cert_Declined.size(); i++) {
                TempStr = PatternContent.matcher(DLVTemplate4Certificate).replaceAll(Cert_Declined.get(i).toString());
                Cert_Declined.setElementAt(TempStr,i);
           }


            //Writing Access Query ...
           FirstIndex = this.Request.ServiceRequest.indexOf(InternalDelimiter);
           if (FirstIndex == this.Request.ServiceRequest.lastIndexOf(InternalDelimiter))
               IsRequestForCertificate = true;
           else
               IsRequestForCertificate = false;
            
                
           String requestAsFact = new String();
           requestAsFact = PatternContent.matcher(DLVTemplate4Service).replaceAll(ServiceRequest) + ".";
                
            //writing the service request as a fact in a tmp file.
            requestAsFactFile = new File(WorkingDirectory  + OSFileDelimiter + "_requestAsFact_" + FileStr);
            FileOutputStream requestAsFactFileStream = new FileOutputStream(requestAsFactFile);
            requestAsFactFileStream.write(requestAsFact.getBytes());
            requestAsFactFileStream.close();
            
            
            if (this.Request.IsService) 
            {
                deductionQueryStream.write((PatternContent.matcher(DLVTemplate4Service).replaceAll(this.Request.ServiceRequest) + "?").getBytes());
            } 
            else 
            {
                if (IsRequestForCertificate)
                {
                    deductionQueryStream.write((PatternContent.matcher(DLVTemplate4Certificate).replaceAll(this.Request.ServiceRequest) + "?").getBytes());
                }
                else
                {       
                    deductionQueryStream.write((PatternContent.matcher(DLVTemplate4Credential).replaceAll(this.Request.ServiceRequest) + "?").getBytes());
                }
            }

            deductionPositiveFactsStream.close();
            deductionQueryStream.close();
            deductionPositiveFactsFileWithAnyValueTermCredsStream.close();
            
            // StepwisePolicy preparation ... [prefixing all heads of rules]
            
                  String  ReadLine                          = "",
                          AllPolicyInteractionRulesAndFacts = "";
                  
                  //!!! Below in this function PolicyInteraction refers to the DisclosurePolicy
                          
                  Pattern pattOneDot = Pattern.compile("\\."),
                          pattRule   = Pattern.compile(":-");
                  
                  // Compiling all rules and facts from Interaction Policy Files in one place
                  String[] OneStepReleaseFiles = PatternWhiteSpaces.split(ComputeAbduciblesPolicyFiles);
                  for(int i=0; i < OneStepReleaseFiles.length; i++)
                  {
                      BufferedReader currentlyOpenFileReader = new BufferedReader( new FileReader(OneStepReleaseFiles[i]) );
                      // To be checked whether it reads end of a file if it does not finished with '\n'!!!!!
                      while((ReadLine = currentlyOpenFileReader.readLine()) != null)
                      {
                          if(ReadLine.length()==0 || ReadLine.startsWith("%")) continue;
                          AllPolicyInteractionRulesAndFacts += ReadLine;
                      }
                      currentlyOpenFileReader.close();
                      
                  }
                  //remove white spaces from all rules so that we can easily compare credentials.
                  AllPolicyInteractionRulesAndFacts = replacePatStr2Str("\\s+",AllPolicyInteractionRulesAndFacts,"");
                                    
                  //Preparing Cred patterns
                  String pattCredStr = PatternRightBrace.matcher(DLVTemplate4Credential).replaceAll(PatternStr2RightBraceStr);    
                  pattCredStr = PatternLeftBrace.matcher(pattCredStr).replaceAll(PatternStr2LeftBraceStr);    
                  pattCredStr = PatternContent.matcher(pattCredStr).replaceAll(".+?");

                  String pattCertStr = PatternRightBrace.matcher(DLVTemplate4Certificate).replaceAll(PatternStr2RightBraceStr);    
                  pattCertStr = PatternLeftBrace.matcher(pattCertStr).replaceAll(PatternStr2LeftBraceStr);    
                  pattCertStr = PatternContent.matcher(pattCertStr).replaceAll(".+?");
                  
                  Pattern pattCred = Pattern.compile(pattCredStr),
                          pattCert = Pattern.compile(pattCertStr);

                  
                  
                  // Spliting entire content into separate facts and rules ...
                  String[] allRulesAndFacts = pattOneDot.split(AllPolicyInteractionRulesAndFacts);
                  for(int i=0; i < allRulesAndFacts.length ; i++)
                  {
                      if(allRulesAndFacts[i].length() == 0) continue;
                      Matcher matcherRule = pattRule.matcher(allRulesAndFacts[i]);
                      Matcher matcherCred;
                      if(matcherRule.find())
                      {
                          String[] HeadAndBodyOfARule = pattRule.split(allRulesAndFacts[i]);
                          matcherCred = pattCred.matcher(HeadAndBodyOfARule[0]);
                          if(matcherCred.find())
                          {
                              // Replacing all predicates for credentials with the prefix/extension .....
                              // Essentially, "$0" tells to the pattern to include the content of the matched predicate
                              HeadAndBodyOfARule[0] = matcherCred.replaceAll(DLVOneStepPrefix4Credentials + "$0");
                          }
                          matcherCred = pattCert.matcher(HeadAndBodyOfARule[0]);
                          if(matcherCred.find())
                          {
                              // Replacing all predicates for certificates with the prefix/extension .....
                              HeadAndBodyOfARule[0] = matcherCred.replaceAll(DLVOneStepPrefix4Credentials + "$0");
                          }
                          //mske the (new) rule back
                          allRulesAndFacts[i] = HeadAndBodyOfARule[0] + ":-" + HeadAndBodyOfARule[1] + ".";
                      }
                      else  // replace a fact if matching
                      {
                          matcherCred = pattCred.matcher(allRulesAndFacts[i]);
                          if(matcherCred.find())
                          {
                              // Replacing all predicates for credentials with the prefix/extension .....
                              // Essentially, "$0" tells to the pattern to include the content of the matched predicate
                              allRulesAndFacts[i] = matcherCred.replaceAll(DLVOneStepPrefix4Credentials + "$0");
                          }
                          matcherCred = pattCert.matcher(allRulesAndFacts[i]);
                          if(matcherCred.find())
                          {
                              // Replacing all predicates for certificates with the prefix/extension .....
                              allRulesAndFacts[i] = matcherCred.replaceAll(DLVOneStepPrefix4Credentials + "$0");
                          }

                          //making the (new) predicates as a fact.
                          allRulesAndFacts[i] += ".";
                      }
                  }

                  //Add back in a String to write down in a file
                  AllPolicyInteractionRulesAndFacts = "";
                  for(int i=0; i < allRulesAndFacts.length ; i++)
                  {
                      AllPolicyInteractionRulesAndFacts += allRulesAndFacts[i];
                  }
                                               
                  //Storing the StepwiseDisclosure policy with rewritten head of rules
                  stepwiseDisclosurePolicy = new File(WorkingDirectory  + OSFileDelimiter + "_stepwiseDisclosurePolicy_" + FileStr);
                  FileOutputStream stepwiseDisclosurePolicyFileStream = new FileOutputStream(stepwiseDisclosurePolicy);
                  stepwiseDisclosurePolicyFileStream.write((requestAsFact+AllPolicyInteractionRulesAndFacts).getBytes());
                  stepwiseDisclosurePolicyFileStream.close();
                  
                  //Storing the StepwiseDisclosurePolicy AND negative complementary rules TOGETHER 
                  stepwiseDisclosurePolicyWithComplementRules = new File(WorkingDirectory  + OSFileDelimiter + "_stepwiseDisclosurePolicyWithComplementRules_" + FileStr);
                  FileOutputStream stepwiseDisclosurePolicyWithComplementRulesFileStream = new FileOutputStream(stepwiseDisclosurePolicyWithComplementRules);
                  //writing the declined credentials as facts with a special prefix
                  for(int j=0; j < Cred_Declined.size(); j++)
                  {
                      String tmp = PatternContent.matcher(DLVTemplate4Credential).replaceAll(Cred_Declined.get(j).toString());
                      tmp = Prefix4DeclinedCred + DLVOneStepPrefix4Credentials + tmp + ".";
                      stepwiseDisclosurePolicyWithComplementRulesFileStream.write(tmp.getBytes());
                  }
                  for(int j=0; j < Cert_Declined.size(); j++)
                  {
                      //String tmp = PatternContent.matcher(DLVTemplate4Credential).replaceAll(Cert_Declined.get(j).toString());
                      String tmp = Prefix4DeclinedCred + DLVOneStepPrefix4Credentials + Cert_Declined.get(j) + ".";
                      stepwiseDisclosurePolicyWithComplementRulesFileStream.write(tmp.getBytes());
                  }
                  //writing all (new) rules of the disclosure policy and the service request
                  stepwiseDisclosurePolicyWithComplementRulesFileStream.write((requestAsFact+AllPolicyInteractionRulesAndFacts).getBytes());
                 
                  //writing the special two rules for deriving credentials not chained by the declined ones
                  stepwiseDisclosurePolicyWithComplementRulesFileStream.write((Rule4NotChainCred).getBytes());
                  stepwiseDisclosurePolicyWithComplementRulesFileStream.write((Rule4NotChainCert).getBytes());
                  stepwiseDisclosurePolicyWithComplementRulesFileStream.close();
                  // End of StepwisePolicy preparation
            
                  resSetUp.setIsSuccessful(true);
                  resSetUp.setOperationResult("Successfully executed SetUp function!");

                } catch(Exception e) {
                        resSetUp.setIsSuccessful(false);
                        resSetUp.setOperationResult(e.toString());
                }
                return resSetUp;
       }

        /**
         * Computes if an access policy together with a set of presented credentials grants a request.
         * It first checks if the access policy is consistent according to the client's presented credentials
         * and then, it checks if the request is in a stable model of the access policy.
         * We use 'cautious' reasoning for computing if the request is in all stable models of the access policy.
         * It is important to note that 'brave' reasoning (consult DLV reasoner for details) my result in inconsistent
         * interactions of missing credentials within a wrong/different stable model of the access policy.
         * <p>We notice that abduction reasoning of DLV (to be fully confirmed) uses 'brave' reasoning for
         * computing missing sets of credentials, and that's why the access policy is also to be a stratified logic program.
         * In this way you guarantee that abduction and deduction in the interactive access control will have consistent
         * behavior. <p>One should experiment if having an access policy with more than one stable models the abduction
         * behavior is consistent with respect to a given service request, which appears in one of the stable
         * models of the access policy. Hope, it is clear.
         */
        private ResultOfOperation doDeduction()
        {
               ResultOfOperation resdoDeduction = new ResultOfOperation();
               try
               {
                   // Executing DLV system ...
                   String ExecStrConsistency = DLVDirectory + OSFileDelimiter + DLVExecutableName     
                       + " -silent" + " " + deductionPositiveFactsFile.getPath() + " ";
                   if (this.Request.IsService)
                       ExecStrConsistency      += DeductionServicePolicyFiles;
                   else
                       ExecStrConsistency      += requestAsFactFile.getPath() + " " + DeductionCredentialPolicyFiles;

                   // Executing DLV system. Checking for consistency ...
                   Process DLVDeductionProcess = Runtime.getRuntime().exec(ExecStrConsistency);
                   // Preparing the output of the dlv execution ...
                   BufferedReader DLVDeductionProcessOutput = new BufferedReader(new InputStreamReader(DLVDeductionProcess.getInputStream()));
                   // if a policy is inconsistent (NO MODEL EXISTS) ...
                   String ConsistencyCheckOutput = DLVDeductionProcessOutput.readLine();
                   if( (ConsistencyCheckOutput == null) || (ConsistencyCheckOutput.trim().compareTo("")==0) )
                   {
                       resdoDeduction.setIsSuccessful(true);
                       resdoDeduction.setOperationResult(new Integer(0)); // Denying service
                       return resdoDeduction;
                   }

                   String ExecStr = DLVDirectory + OSFileDelimiter + DLVExecutableName     +
                   " " + DLVDeductionCommandPromptOptions     +
                   " " + deductionPositiveFactsFile.getPath() +
                   " " + deductionQueryFile.getPath()         +
                   " ";
                   if (this.Request.IsService)
                       ExecStr += DeductionServicePolicyFiles;
                   else
                       ExecStr += requestAsFactFile.getPath() + " " + DeductionCredentialPolicyFiles;

                   // Executing DLV system ...
                   DLVDeductionProcess = Runtime.getRuntime().exec(ExecStr);
                   // Preparing the output of the dlv execution ...
                   DLVDeductionProcessOutput = new BufferedReader(new InputStreamReader(DLVDeductionProcess.getInputStream()));

                   Pattern patternCredentialsDelimiter = Pattern.compile("\\s*" + CredentialsDelimiter + "\\s*");
                   String[] keywords = patternCredentialsDelimiter.split(DLVKeywords4SUccessfulDeduction);

                   int countSucessKeywords = 0;
                   boolean grantAccess = false;
                   String ReadLine;
                   while((ReadLine = DLVDeductionProcessOutput.readLine()) != null)
                   {
                        for(int i=0; i< keywords.length; i++)
                        {
                             Pattern pat = Pattern.compile(keywords[i]);
                             Matcher matcher = pat.matcher(ReadLine);
                             if(matcher.find())
                             {
                                 countSucessKeywords++;
                                 if(countSucessKeywords == keywords.length)
                                     grantAccess = true;
                             }
                        }
                        if(grantAccess) break;
                   }

                   resdoDeduction.setIsSuccessful(true);
                   if(grantAccess)
                      resdoDeduction.setOperationResult(new Integer(1)); // Granting service
                   else
                      resdoDeduction.setOperationResult(new Integer(0)); // Denying service
                   
                   return resdoDeduction;
              }
              catch(Exception e)
              {
                  resdoDeduction.setIsSuccessful(false);
                   resdoDeduction.setOperationResult(e.toString());
                   return resdoDeduction;
              }
       }

        /**
         * Computes the set of disclosable credentials in accordance with the
         * interactive access control algorithm, implemented in function computeAccessDecision().
         * The function assumes that the disclosure policy is a stratified logic programm, i.e., it has unique
         * stable model. So, the function takes ONLY the FIRST stable model of the disclosure policy union
         * the presented credentials. 
         */
        private ResultOfOperation computeAbducibles()
        {
              ResultOfOperation resComputeAbducibles = new ResultOfOperation();
              LinkedList abducibleModels = new LinkedList();
              String ReadLine             = "",
                     declinedCredPattern  = "",
                     presentedCredPattern = "";
              try
              {
                   // Executing DLV system ...
                   Process DLVAbduciblesProcess = Runtime.getRuntime().exec( DLVDirectory + OSFileDelimiter + DLVExecutableName         +
                                                                             " " + DLVComputeAbduciblesCommandPromptOptions +
                                                                             " " + deductionPositiveFactsFile.getPath()     +
                                                                             " " + stepwiseDisclosurePolicyWithComplementRules.getPath()
                                                                           );
                   // Preparing the output of the dlv execution ...
                   BufferedReader DLVAbduciblesProcessOutput = new BufferedReader(new InputStreamReader(DLVAbduciblesProcess.getInputStream()));

                   // checking if there is an answer set ...
                   ReadLine = DLVAbduciblesProcessOutput.readLine();
                   if(ReadLine == null || ReadLine.equals("") || ReadLine.equals("{}"))
                   {
                         resComputeAbducibles.setIsSuccessful(true);
                         resComputeAbducibles.setOperationResult(null);
                         return resComputeAbducibles;
                   }

                   // at this point at least one answer set exists ....
                   // Preparing abducibles in a form suitable for the abduction computation ...
                   // transforming it in a form "predicate1(). predicate2(). ..."
                   // in this case they are ready to be saved derectly in a file.hyp for the real abductive process.
                   ReadLine = replacePatStr2Str("\\{", ReadLine, "");
                   ReadLine = replacePatStr2Str("\\}", ReadLine, ".");
                   ReadLine = replacePatStr2Str("\\)\\,", ReadLine, ").");
                   ReadLine = replacePatStr2Str("\\,\\s", ReadLine, ".");

                   // Removing presented credentials and certificates from the set of abducibles...
            for(int i=0; i < Cred_Presented.size(); i++)
            {
               presentedCredPattern = PatternContent.matcher(DLVExtendedTemplate4Credential).replaceAll(Cred_Presented.get(i).toString() + "\\,[0-9]+?");
               presentedCredPattern = PatternRightBrace.matcher(presentedCredPattern).replaceAll(PatternStr2RightBraceStr);
               presentedCredPattern = PatternLeftBrace.matcher(presentedCredPattern).replaceAll(PatternStr2LeftBraceStr);
               ReadLine = replacePatStr2Str(presentedCredPattern+"\\.", ReadLine, "");
            }
            for(int i=0; i < Cert_Presented.size(); i++)
            {
                presentedCredPattern = PatternRightBrace.matcher(Cert_Presented.get(i).toString()).replaceAll(PatternStr2RightBraceStr);
                presentedCredPattern = PatternLeftBrace.matcher(presentedCredPattern).replaceAll(PatternStr2LeftBraceStr);
                ReadLine = replacePatStr2Str(presentedCredPattern+"\\.", ReadLine, "");
            }

            // Removing declined credentials and certificates from the set of abducibles...
            for(int i=0; i < Cred_Declined.size(); i++)
            {
                declinedCredPattern = PatternContent.matcher(DLVExtendedTemplate4Credential).replaceAll(Cred_Declined.get(i).toString() + "\\,[0-9]+?");
                declinedCredPattern = PatternRightBrace.matcher(declinedCredPattern).replaceAll(PatternStr2RightBraceStr);
                declinedCredPattern = PatternLeftBrace.matcher(declinedCredPattern).replaceAll(PatternStr2LeftBraceStr);
                ReadLine = replacePatStr2Str(declinedCredPattern+"\\.", ReadLine, "");
            }

            for(int i=0; i < Cert_Declined.size(); i++)
            {
                declinedCredPattern = PatternRightBrace.matcher(Cert_Declined.get(i).toString()).replaceAll(PatternStr2RightBraceStr);
                declinedCredPattern = PatternLeftBrace.matcher(declinedCredPattern).replaceAll(PatternStr2LeftBraceStr);
                ReadLine = replacePatStr2Str(declinedCredPattern+"\\.", ReadLine, "");
            }

           // [NEW] Removing all presented credentials and certificates with anyValueTerm inside
            for(int i=0; i < Cred_Presented_With_AnyValueTerm.size(); i++)
            {
               presentedCredPattern = PatternContent.matcher(DLVExtendedTemplate4Credential).replaceAll(Cred_Presented_With_AnyValueTerm.get(i).toString() + "\\,[0-9]+?");
               presentedCredPattern = PatternRightBrace.matcher(presentedCredPattern).replaceAll(PatternStr2RightBraceStr);
               presentedCredPattern = PatternLeftBrace.matcher(presentedCredPattern).replaceAll(PatternStr2LeftBraceStr);
               ReadLine = replacePatStr2Str(presentedCredPattern+"\\.", ReadLine, "");
            }
            for(int i=0; i < Cert_Presented_With_AnyValueTerm.size(); i++)
            {
                presentedCredPattern = PatternContent.matcher(DLVTemplate4Certificate).replaceAll(Cert_Presented_With_AnyValueTerm.get(i).toString());
                presentedCredPattern = PatternRightBrace.matcher(presentedCredPattern).replaceAll(PatternStr2RightBraceStr);
                presentedCredPattern = PatternLeftBrace.matcher(presentedCredPattern).replaceAll(PatternStr2LeftBraceStr);
                ReadLine = replacePatStr2Str(presentedCredPattern+"\\.", ReadLine, "");
            }
                        
            // Abducible Model:  ReadLine
            // adding them as a separate set of abducibles ...

            abducibleModels.add(ReadLine);

            resComputeAbducibles.setIsSuccessful(true);
            resComputeAbducibles.setOperationResult(abducibleModels);
            return resComputeAbducibles;
           }
           catch(Exception e)
           {
               resComputeAbducibles.setIsSuccessful(false);
               resComputeAbducibles.setOperationResult(e.toString());
               return resComputeAbducibles;
           }
        }

        
        
    /**
     * Extracts a missing set of credentials to a form of the negotiation layer, that is, the function strips down
     * credential's and certificate's notations and returns only their values. For example, "cred(john_couk,social_worker,california_ca)" is
     * transformed to "john_couk,social_worker,california_ca".
     */
    private ResultOfOperation ExtractMinimalCredentials(String[] inputCredentials)
    {
            ResultOfOperation resextCredContModel = new ResultOfOperation();
            Vector                CertCont = new Vector(),
                                  CredCont = new Vector();

            try
            {
            String newCredPattStr = PatternLeftBrace.matcher(DLVExtendedTemplate4Credential).replaceAll(PatternStr2LeftBraceStr);
            newCredPattStr = PatternRightBrace.matcher(newCredPattStr).replaceAll(PatternStr2RightBraceStr);
            newCredPattStr = PatternContent.matcher(newCredPattStr).replaceAll("(.+?)\\,[0-9]+?");

            String newCertPattStr = PatternLeftBrace.matcher(DLVTemplate4Certificate).replaceAll(PatternStr2LeftBraceStr);
            newCertPattStr = PatternRightBrace.matcher(newCertPattStr).replaceAll(PatternStr2RightBraceStr);
            newCertPattStr = PatternContent.matcher(newCertPattStr).replaceAll("(.+?)");

            Pattern PattCred = Pattern.compile(newCredPattStr);
            Matcher credMatcher;

            Pattern PattCert = Pattern.compile(newCertPattStr);
            Matcher certMatcher;

            for(int i=0; i<inputCredentials.length; i++)
            {
                  credMatcher = PattCred.matcher(inputCredentials[i]);
                  if(credMatcher.find()) // current credential is certificate
                          CredCont.add(credMatcher.group(1));

                  certMatcher = PattCert.matcher(inputCredentials[i]);
                  if (certMatcher.find())
                          CertCont.add(certMatcher.group(1));

            }

            CertCont.addAll(CredCont);

            resextCredContModel.setIsSuccessful(true);
            resextCredContModel.setOperationResult(CertCont);
            return resextCredContModel;
            } 
            catch(Exception e) 
            {
                resextCredContModel.setIsSuccessful(false);
                resextCredContModel.setOperationResult(e.toString());
                return resextCredContModel;
            }
    }


        /**
         * Computes a minimal set of missing credentials. It is used by the
         * core access control algorithm, in function computeAccessDecision();
         * It has an input list of sets of disclosable credentials, BUT only the first set is actually
         * used. Initially, we wanted to compute a set of missing credentials among more than one sets of
         * disclosable credentials, but it was against the theory, that is, the disclosure policy is
         * to be a stratified logic program and, as such, it has only one stable model -- one set
         * of disclosable credentials.
         */
        private ResultOfOperation doAbduction(LinkedList abducibleSets)
        {
            ResultOfOperation resdoAbduction = new ResultOfOperation();
             try
             {
                // writing abducibles into a file with extension .obs ...
                abductionObservationsFile = new File(WorkingDirectory  + OSFileDelimiter + "_abductionObservations_" + FileStr + observationsFileExt);
                FileOutputStream abductionObservationsStream = new FileOutputStream(abductionObservationsFile);

                if (this.Request.IsService)
                     abductionObservationsStream.write((PatternContent.matcher(DLVTemplate4Service).replaceAll(this.Request.ServiceRequest) + ".").getBytes());
                else 
                {
                     if (IsRequestForCertificate)
                             abductionObservationsStream.write((PatternContent.matcher(DLVTemplate4Certificate).replaceAll(this.Request.ServiceRequest) + ".").getBytes());
                     else
                             abductionObservationsStream.write((PatternContent.matcher(DLVTemplate4Credential).replaceAll(this.Request.ServiceRequest) + ".").getBytes());
                }
                abductionObservationsStream.close();

                //Initializing patterns ...
                String pattRoleWeightStr = PatternLeftBrace.matcher(DLVExtendedTemplate4Credential).replaceAll(PatternStr2LeftBraceStr);
                pattRoleWeightStr = PatternRightBrace.matcher(pattRoleWeightStr).replaceAll(PatternStr2RightBraceStr);
                pattRoleWeightStr = PatternContent.matcher(pattRoleWeightStr).replaceAll(".+?\\,([0-9]+?)");
                Pattern pattExtractRoleWeightFromExtendedCredential = Pattern.compile(pattRoleWeightStr);

                Pattern patt4CountingSetCardinality = Pattern.compile("\\s+?");

                int beginIndex = "Diagnosis: ".length();
                String AbductionHypFileStr = WorkingDirectory  + OSFileDelimiter + "_abductionHypotheses_" + FileStr + hypothesesFileExt;
                abductionHypothesesFile = new File(AbductionHypFileStr);
                String DLVStr = DLVDirectory + OSFileDelimiter + DLVExecutableName +
                                    " " + DLVAbductionCommandPromptOptions     +
                                    // use CP U CP_anyvalue term
                                    " " + deductionPositiveFactsFileWithAnyValueTermCreds.getPath() + 
                                    " " + abductionObservationsFile.getPath()  +
                                    " " + AbductionHypFileStr;
                
                // Below we check if to add a service request as a fact in a credential policy
                if(this.Request.IsService)
                     DLVStr += " " + AbductionPolicyFiles;
                else
                     DLVStr += " " + requestAsFactFile.getPath() + " " + AbductionCredentialPolicyFiles;

                String ReadLine = "";
                // writing hypotheses into a file with extension .hyp ...
                // WE TAKE ONLY THE FIRST SET OF HYPOTHESES, ACTUALLY THE ONLY ONE ASSUMED BY THE THEORY
                // WE HAVE A LINKEDLIST AS INPUT BUT ONLY GET(0) IS NECESSARY.
                FileOutputStream abductionHypothesesStream = new FileOutputStream(abductionHypothesesFile);
                abductionHypothesesStream.write(((String) abducibleSets.get(0)).getBytes());
                abductionHypothesesStream.close();

                // Executing DLV system ...
                Process DLVAbductionProcess = Runtime.getRuntime().exec(DLVStr);
                // Preparing the output of the dlv execution ...
                BufferedReader DLVAbductionProcessOutput = new BufferedReader(new InputStreamReader(DLVAbductionProcess.getInputStream()));

                ReadLine = DLVAbductionProcessOutput.readLine();

                if(ReadLine == null || ReadLine.equals("") || ReadLine.equals("Diagnosis:"))
                {
                  resdoAbduction.setIsSuccessful(true);
                  resdoAbduction.setOperationResult(null);
                  return resdoAbduction;
                }

                // at least one answer set exists ....
                //Preparing ReadLine for computing ...
                ReadLine = ReadLine.substring(beginIndex);

                int    minSetCardinality     = 0,
                       currentSetCardinality = 0,
                       minSumOfWeights       = 0,
                       currentSumOfWeights   = 0;
                //points to the minimal set
                String minSetStr = new String(ReadLine);

                // Counting Set Cardinality ...
                Matcher matchSetCardinality = patt4CountingSetCardinality.matcher(ReadLine);
                while(matchSetCardinality.find())
                     minSetCardinality++;
                minSetCardinality++; // real cardinality is + 1

                // Adding Attr Weight Value ...
                Matcher matchRoleWeight = pattExtractRoleWeightFromExtendedCredential.matcher(ReadLine);
                while(matchRoleWeight.find())
                     minSumOfWeights += Integer.parseInt(matchRoleWeight.group(1));

                while((ReadLine = DLVAbductionProcessOutput.readLine()) != null)
                {
                    ReadLine = ReadLine.substring(beginIndex);
                    currentSetCardinality = 0;
                    currentSumOfWeights   = 0;

                    // Counting the current set cardinality
                    matchSetCardinality = patt4CountingSetCardinality.matcher(ReadLine);
                    while(matchSetCardinality.find())
                        currentSetCardinality++;
                    currentSetCardinality++; // real cardinality is + 1

                    // Counting the current sum of weights
                    matchRoleWeight = pattExtractRoleWeightFromExtendedCredential.matcher(ReadLine);
                    while(matchRoleWeight.find())
                        currentSumOfWeights += Integer.parseInt(matchRoleWeight.group(1));

                    if(CriterionRoleMinimalitySetCardinality)
                    {
                        // if current set is least privileged that the minimal one
                        if( currentSumOfWeights < minSumOfWeights )
                        {
                            minSumOfWeights = currentSumOfWeights;
                            minSetCardinality = currentSetCardinality;
                            minSetStr = ReadLine;
                        }
                        // if current set is equally privileged that the minimal one
                        if(currentSumOfWeights == minSumOfWeights)
                        {
                            if( currentSetCardinality < minSetCardinality )
                            {
                                minSumOfWeights = currentSumOfWeights;
                                minSetCardinality = currentSetCardinality;
                                minSetStr = ReadLine;
                            }
                        }
                    }
                    else
                    {
                        if( currentSetCardinality < minSetCardinality )
                        {
                            minSumOfWeights = currentSumOfWeights;
                            minSetCardinality = currentSetCardinality;
                            minSetStr = ReadLine;
                        }
                        if( currentSetCardinality == minSetCardinality )
                        {
                            // if current set is least privileged that the minimal one
                            if( currentSumOfWeights < minSumOfWeights )
                            {
                                minSumOfWeights = currentSumOfWeights;
                                minSetCardinality = currentSetCardinality;
                                minSetStr = ReadLine;
                            }
                        }
                    }

                }// end of while

                //return result of minimal set
                String[] st = PatternWhiteSpaces.split(minSetStr);
                return ExtractMinimalCredentials(st);
             }
             catch(Exception e)
             {
                  resdoAbduction.setIsSuccessful(false);
                  resdoAbduction.setOperationResult(e.toString());
                  return resdoAbduction;
             }

        }


        /**
         * Performs the stepwise computing of missing credentials.
         * Important, before to run this function you have to use the SetUp() function of this class.
         * <p>The function takes as input a set of missing credentials and returns a next step
         * of credentials (among the missing ones) to be asked to an opponent.
         * Similarly in computing to the immediate consequence operator in logic programming.
         * The disclosure policy (a copy of it) is internally manipulated to disclose
         * a next set of credentials. We use internally negation as failure on a set of credentials
         * to make the computing. 
         * <p>Important, we guarantee that if the disclosure policy is a stratified logic program,
         * the internal transformations preserve the disclosure policy stratified. That is, we use
         * negation as failure on facts internally generated and added to the policy, and preserve the single stable model
         * of the policy.
         */
        public ResultOfOperation doStepwiseDisclosure(Vector MissingCreds)
        {
            ResultOfOperation resStepwiseDisclosure = new ResultOfOperation();
	    try
            {
                  String ReadLine = "";
                  String InternalDelimiter = ",";
                  String replaceStr = "";
                  String TempStr;
                  int FirstIndex = -1;
                  
                  if(MissingCreds == null)
                  {
                        deleteTempFiles();
                        resStepwiseDisclosure.setIsSuccessful(true);
                        resStepwiseDisclosure.setOperationResult(null);
                        return resStepwiseDisclosure;                    
                  }
                  
                  Vector CM = new Vector();
                  CM.addAll((Vector) MissingCreds.clone());
                  
                      for (int i = 0; i < CM.size(); i++)
                      {
                          FirstIndex = CM.get(i).toString().indexOf(InternalDelimiter);
                          
                          if (FirstIndex == -1)
                              throw new Exception("Could not identify a valid credential format in CM of the stepwise function.");
                          if (FirstIndex == CM.get(i).toString().lastIndexOf(InternalDelimiter))
                          {
                              TempStr = PatternWhiteSpaces.matcher(CM.get(i).toString()).replaceAll(replaceStr);
                              TempStr = PatternContent.matcher(DLVTemplate4Certificate).replaceAll(TempStr);
                              CM.setElementAt(TempStr,i);
                          }                              
                          else //Identified credential by text,text,text
                          {
                              TempStr = PatternWhiteSpaces.matcher(CM.get(i).toString()).replaceAll(replaceStr);
                              TempStr = PatternContent.matcher(DLVTemplate4Credential).replaceAll(TempStr);
                              CM.setElementAt(TempStr,i);
                          }
                      }
                  
                  // Executing DLV system ...
		  Process DLVOneStepDeductionProcess = Runtime.getRuntime().exec(DLVDirectory + OSFileDelimiter + DLVExecutableName        + 
                                                                                 " " + DLVOneStepDeductionCommandPromptOptions + 
                                                                                 " " + deductionPositiveFactsFile.getPath()    +
                                                                                 " " + stepwiseDisclosurePolicy.getPath()           
                                                                                 );			
	          // Preparing the output of the dlv execution ...
		  BufferedReader DLVOneStepDeductionProcessOutput = new BufferedReader(new InputStreamReader(DLVOneStepDeductionProcess.getInputStream()));

                  String presentedCredPattern,
                         declinedCredPattern;
                  //Read the first model from DLV. If more than one , first *unstable behavior by theory*, 
                  //second we take only the first one
                  ReadLine = DLVOneStepDeductionProcessOutput.readLine();
                  if(ReadLine == null)
                  {
                        deleteTempFiles();
                        resStepwiseDisclosure.setIsSuccessful(true);
                        resStepwiseDisclosure.setOperationResult(null);
                        return resStepwiseDisclosure;                    
                  }

                  ReadLine = replacePatStr2Str("\\{", ReadLine, "");
                  ReadLine = replacePatStr2Str("\\}", ReadLine, ".");
                  ReadLine = replacePatStr2Str("\\)\\,", ReadLine, ").");
                  ReadLine = replacePatStr2Str("\\,\\s", ReadLine, ".");
                      

                  // Removing the one step prefix used for deduction process ...
                  ReadLine = replacePatStr2Str(DLVOneStepPrefix4Credentials, ReadLine, "");
                  // Raw CD1: ReadLine
                  // Removing presented credentials and certificates from the set of abducibles...

                  for(int i=0; i < Cred_Presented.size(); i++)
                      {
                           presentedCredPattern = PatternContent.matcher(DLVExtendedTemplate4Credential).replaceAll(Cred_Presented.get(i).toString() + "\\,[0-9]+?");
                           presentedCredPattern = PatternRightBrace.matcher(presentedCredPattern).replaceAll(PatternStr2RightBraceStr);
                           presentedCredPattern = PatternLeftBrace.matcher(presentedCredPattern).replaceAll(PatternStr2LeftBraceStr);
                           ReadLine = replacePatStr2Str(presentedCredPattern+"\\.", ReadLine, "");
                      }

                  for(int i=0; i < Cert_Presented.size(); i++)
                      {
                          presentedCredPattern = PatternRightBrace.matcher(Cert_Presented.get(i).toString()).replaceAll(PatternStr2RightBraceStr);
                          presentedCredPattern = PatternLeftBrace.matcher(presentedCredPattern).replaceAll(PatternStr2LeftBraceStr);
                          ReadLine = replacePatStr2Str(presentedCredPattern+"\\.", ReadLine, "");
                      }

                      // Removing declined credentials and certificates from the set of abducibles...
                  for(int i=0; i < Cred_Declined.size(); i++)
                      {
                            declinedCredPattern = PatternContent.matcher(DLVExtendedTemplate4Credential).replaceAll(Cred_Declined.get(i).toString() + "\\,[0-9]+?");
                            declinedCredPattern = PatternRightBrace.matcher(declinedCredPattern).replaceAll(PatternStr2RightBraceStr);
                            declinedCredPattern = PatternLeftBrace.matcher(declinedCredPattern).replaceAll(PatternStr2LeftBraceStr);
                            ReadLine = replacePatStr2Str(declinedCredPattern+"\\.", ReadLine, "");
                      }

                  for(int i=0; i < Cert_Declined.size(); i++)
                      {
                            declinedCredPattern = PatternRightBrace.matcher(Cert_Declined.get(i).toString()).replaceAll(PatternStr2RightBraceStr);
                            declinedCredPattern = PatternLeftBrace.matcher(declinedCredPattern).replaceAll(PatternStr2LeftBraceStr);
                            ReadLine = replacePatStr2Str(declinedCredPattern+"\\.", ReadLine, "");
                      }

                       // [NEW] Removing all presented credentials and certificates with anyValueTerm inside
                  for(int i=0; i < Cred_Presented_With_AnyValueTerm.size(); i++)
                       {
                           presentedCredPattern = PatternContent.matcher(DLVExtendedTemplate4Credential).replaceAll(Cred_Presented_With_AnyValueTerm.get(i).toString() + "\\,[0-9]+?");
                           presentedCredPattern = PatternRightBrace.matcher(presentedCredPattern).replaceAll(PatternStr2RightBraceStr);
                           presentedCredPattern = PatternLeftBrace.matcher(presentedCredPattern).replaceAll(PatternStr2LeftBraceStr);
                           ReadLine = replacePatStr2Str(presentedCredPattern+"\\.", ReadLine, "");
                       }

                  for(int i=0; i < Cert_Presented_With_AnyValueTerm.size(); i++)
                       {
                            presentedCredPattern = PatternContent.matcher(DLVTemplate4Certificate).replaceAll(Cert_Presented_With_AnyValueTerm.get(i).toString());
                            presentedCredPattern = PatternRightBrace.matcher(presentedCredPattern).replaceAll(PatternStr2RightBraceStr);
                            presentedCredPattern = PatternLeftBrace.matcher(presentedCredPattern).replaceAll(PatternStr2LeftBraceStr);
                            ReadLine = replacePatStr2Str(presentedCredPattern+"\\.", ReadLine, "");
                       }
                      
                  //CD1\(CN U CP): ReadLine
                  if(ReadLine.length() == 0)
                  {
                        deleteTempFiles();
                        resStepwiseDisclosure.setIsSuccessful(true);
                        resStepwiseDisclosure.setOperationResult(null);
                        return resStepwiseDisclosure;                    
                  }
                  ReadLine = replacePatStr2Str("\\s+",ReadLine,"");
                  if(ReadLine.length() == 0)
                  {
                        deleteTempFiles();
                        resStepwiseDisclosure.setIsSuccessful(true);
                        resStepwiseDisclosure.setOperationResult(null);
                        return resStepwiseDisclosure;                    
                  }

                  Pattern pattSplitOutput = Pattern.compile("\\.");
                  String[] OneStepCredentials = pattSplitOutput.split(ReadLine);

                  
                  //Storing the StepwiseDisclosure for one step abduction
                  stepwiseDisclosurePolicyForOneStep = new File(WorkingDirectory  + OSFileDelimiter + "_stepwiseDisclosurePolicyForOneStep_" + FileStr);
                  FileOutputStream stepwiseDisclosurePolicyForOneStepFileStream = new FileOutputStream(stepwiseDisclosurePolicyForOneStep);
//---------
                  for(int j=0; j < Cred_Declined.size(); j++)
                  {
                      String tmp = PatternContent.matcher(DLVTemplate4Credential).replaceAll(Cred_Declined.get(j).toString());
                      tmp = Prefix4DeclinedCred + DLVOneStepPrefix4Credentials + tmp + ".";
                      stepwiseDisclosurePolicyForOneStepFileStream.write(tmp.getBytes());
                  }
                  for(int j=0; j < Cert_Declined.size(); j++)
                  {
                      //String tmp = PatternContent.matcher(DLVTemplate4Credential).replaceAll(Cert_Declined.get(j).toString());
                      String tmp = Prefix4DeclinedCred + DLVOneStepPrefix4Credentials + Cert_Declined.get(j) + ".";
                      stepwiseDisclosurePolicyForOneStepFileStream.write(tmp.getBytes());
                  }
                    

                  String newCredPattStr = PatternLeftBrace.matcher(DLVExtendedTemplate4Credential).replaceAll(PatternStr2LeftBraceStr);
                  newCredPattStr = PatternRightBrace.matcher(newCredPattStr).replaceAll(PatternStr2RightBraceStr);
                  newCredPattStr = PatternContent.matcher(newCredPattStr).replaceAll("(.+?)\\,[0-9]+?");

                  Pattern pattExtCred = Pattern.compile(newCredPattStr);
                  for(int i=0; i < OneStepCredentials.length; i++)
                  {

                      Matcher matchedExtCred = pattExtCred.matcher(OneStepCredentials[i]);
                      String toStore = null;
                      if(matchedExtCred.find())
                          toStore = PatternContent.matcher(DLVTemplate4Credential).replaceAll(matchedExtCred.group(1));
                      else
                          toStore = OneStepCredentials[i];
                      
                      String tmp = Prefix4DeclinedCred + DLVOneStepPrefix4Credentials + toStore + ".";
                      stepwiseDisclosurePolicyForOneStepFileStream.write(tmp.getBytes());
                  }
//----------                  
                  //writing the two rules, already initialized in setup function
                  stepwiseDisclosurePolicyForOneStepFileStream.write(Rule4NotChainCred.getBytes());
                  stepwiseDisclosurePolicyForOneStepFileStream.write(Rule4NotChainCert.getBytes());
                  stepwiseDisclosurePolicyForOneStepFileStream.close();
                  
                  //Storing the Query for abduction 
                  String query;
                  if(CM == null || CM.size() == 0)
                      query = "q.";
                  else
                  {
                      query = "q :- ";
                      for(int i=0; i < CM.size(); i++)
                      {
                          if( i+1 == CM.size())
                              query = query + CM.get(i).toString() + ".";
                          else
                              query = query + CM.get(i).toString() + ",";
                      }
                  }
                  queryOneStep = new File(WorkingDirectory  + OSFileDelimiter + "_queryOneStep_" + FileStr);
                  FileOutputStream queryOneStepFileStream = new FileOutputStream(queryOneStep);
                  queryOneStepFileStream.write((query).getBytes());
                  queryOneStepFileStream.close();

                  oneStepObservations = new File(WorkingDirectory  + OSFileDelimiter + "_oneStepObservations_" + FileStr + observationsFileExt);
                  FileOutputStream oneStepObservationsFileStream = new FileOutputStream(oneStepObservations);
                  oneStepObservationsFileStream.write(("q.").getBytes());
                  oneStepObservationsFileStream.close();
                  
                  oneStepHypotheses = new File(WorkingDirectory  + OSFileDelimiter + "_oneStepHypotheses_" + FileStr + hypothesesFileExt);
                  FileOutputStream oneStepHypothesesFileStream = new FileOutputStream(oneStepHypotheses);
                  oneStepHypothesesFileStream.write((ReadLine).getBytes());
                  oneStepHypothesesFileStream.close();

                  // *** ABDUCTION ***

                  String DLVStr = DLVDirectory + OSFileDelimiter + DLVExecutableName +
                                    " " + DLVAbductionOneStepCommandPromptOptions     +
                                    " " + deductionPositiveFactsFile.getPath() +
                                    " " + queryOneStep.getPath()  +
                                    " " + stepwiseDisclosurePolicy.getPath()  +
                                    " " + stepwiseDisclosurePolicyForOneStep.getPath()  +
                                    " " + OneStepAbductionPolicyFiles + 
                                    " " + oneStepHypotheses.getPath()  +
                                    " " + oneStepObservations.getPath();

                  // Executing DLV system ...
                Process DLVAbductionProcess = Runtime.getRuntime().exec(DLVStr);
                // Preparing the output of the dlv execution ...
                BufferedReader DLVAbductionProcessOutput = new BufferedReader(new InputStreamReader(DLVAbductionProcess.getInputStream()));

                
                Pattern patt4CountingSetCardinality = Pattern.compile("\\s+?");
                String pattRoleWeightStr = PatternLeftBrace.matcher(DLVExtendedTemplate4Credential).replaceAll(PatternStr2LeftBraceStr);
                pattRoleWeightStr = PatternRightBrace.matcher(pattRoleWeightStr).replaceAll(PatternStr2RightBraceStr);
                pattRoleWeightStr = PatternContent.matcher(pattRoleWeightStr).replaceAll(".+?\\,([0-9]+?)");
                Pattern pattExtractRoleWeightFromExtendedCredential = Pattern.compile(pattRoleWeightStr);

                int beginIndex = "Diagnosis: ".length();
                boolean flag4existingSet = false;
                String minSetStr = new String();
                int minSetCardinalityInt  = 0,
                    currentSetCardinality = 0,
                    minSumOfWeights       = 0,
                    currentSumOfWeights   = 0;
                
                //Start enumerating all missing sets ...
                while((ReadLine = DLVAbductionProcessOutput.readLine())!=null)
                {
                    //If the empty set is a solution then return null
                    if(ReadLine.equals("Diagnosis:")) 
                    {
                        deleteTempFiles();
                        resStepwiseDisclosure.setIsSuccessful(true);
                        resStepwiseDisclosure.setOperationResult(null);
                        return resStepwiseDisclosure;                    
                    }
                    //initialize current variables
                    currentSetCardinality = 0;
                    currentSumOfWeights   = 0;
                    
                    //removing "Diagnosis: " string from the readLine
                    ReadLine = ReadLine.substring(beginIndex);

                    //Counting current set cardinality ...
                    Matcher matchSetCardinality = patt4CountingSetCardinality.matcher(ReadLine);
                    while(matchSetCardinality.find())
                        currentSetCardinality++;
                    currentSetCardinality++; // real cardinality is + 1
                    
                    // Adding Attr Weights Value ...
                    Matcher matchRoleWeight = pattExtractRoleWeightFromExtendedCredential.matcher(ReadLine);
                    while(matchRoleWeight.find())
                          currentSumOfWeights += Integer.parseInt(matchRoleWeight.group(1));

                    // Initializing first set as minimal one.
                    if(!flag4existingSet)
                    {
                        minSetCardinalityInt = currentSetCardinality;
                        minSumOfWeights = currentSumOfWeights;
                        minSetStr = ReadLine;
                    }
                    else
                    {
                          if(CriterionRoleMinimalitySetCardinality)
                          {
                              if( currentSumOfWeights < minSumOfWeights )
                              {
                                    minSetCardinalityInt = currentSetCardinality;
                                    minSumOfWeights = currentSumOfWeights;
                                    minSetStr = ReadLine;
                              }
                              if(currentSumOfWeights == minSumOfWeights)
                              {
                                  if( currentSetCardinality < minSetCardinalityInt)
                                  {
                                      minSetCardinalityInt = currentSetCardinality;
                                      minSumOfWeights = currentSumOfWeights;
                                      minSetStr = ReadLine;
                                  }
                              }
                          }
                          else
                          {
                              if( currentSetCardinality < minSetCardinalityInt )
                              {
                                     minSetCardinalityInt = currentSetCardinality;
                                     minSumOfWeights = currentSumOfWeights;
                                     minSetStr = ReadLine;
                              }
                              if(currentSetCardinality == minSetCardinalityInt)
                              {
                                   if(  currentSumOfWeights < minSumOfWeights )
                                   {
                                       minSetCardinalityInt = currentSetCardinality;
                                       minSumOfWeights = currentSumOfWeights;
                                       minSetStr = ReadLine;
                                   }
                              }
                          }
                    }
                    flag4existingSet = true;
                }

                if(!flag4existingSet)// return null as a result
                {
                    deleteTempFiles();
                    resStepwiseDisclosure.setIsSuccessful(true);
                    resStepwiseDisclosure.setOperationResult(null);
                    return resStepwiseDisclosure;                    
                }
                else // return the parsed credential information of the CM1
                {
                    deleteTempFiles();
                    //CM1: minSetCardinalityStr
                    String[] st = PatternWhiteSpaces.split(minSetStr);
                    return ExtractMinimalCredentials(st);
                }


             }
             catch(Exception e)
             {
                 System.err.println("IACCESS: EXCEPTION: BasicAccessControlAlgorithm.doStepwiseDisclosure(): " + e.toString());
                 deleteTempFiles();
                 resStepwiseDisclosure.setIsSuccessful(false);
                 resStepwiseDisclosure.setOperationResult(e.toString());
                 return resStepwiseDisclosure;
             }

         }

        
        
        /** Implements the interactive access control algorithm.
         * It computes grant/deny/MissingCredentials acccording to the
         * interactive algorithm. 
         * Important: before to invoke this function you have to run the SetUp() function of this class.
         * <p><b>1. Deduction:</b> Computes if an access policy together with a set of presented credentials grants a request.
         * It first checks if the access policy is consistent according to the client's presented credentials
         * and then, it checks if the request is in a stable model of the access policy.
         * We use 'cautious' reasoning for computing if the request is in all stable models of the access policy.
         * It is important to note that 'brave' reasoning (consult DLV reasoner for details) my result in inconsistent
         * interactions of missing credentials within a wrong/different stable model of the access policy.
         * <p>We notice that abduction reasoning of DLV (to be fully confirmed) uses 'brave' reasoning for
         * computing missing sets of credentials, and that's why the access policy is also to be a stratified logic program.
         * In this way you guarantee that abduction and deduction in an interactive access control process will have consistent
         * behavior. <p>One should experiment of having an access policy with more than one stable models, and a given service which appears in more than one stable
         * models of the access policy but with different credential requirements, if the abduction behaves as expected.
         * <p><b>2. Disclosable credentials: </b> Computes the set of disclosable credentials in accordance with the
         * interactive access control algorithm.
         * The function assumes that the disclosure policy is a stratified logic programm, i.e., it has unique
         * stable model. So, the function computeAbducibles() takes ONLY the FIRST stable model of the disclosure policy (union
         * the presented credentials).
         * <p><b>3. Abduction: </b> Computes a minimal set of missing credentials.
         * It has an input list of sets of disclosable credentials, BUT , as mentioned, only the first element is actually
         * used. Initially, we wanted to compute a set of missing credentials among more than one sets of
         * disclosable credentials, but it was against the theory, that is, the disclosure policy is
         * to be a stratified logic program and, as such, it has only one stable model -- single set
         * of disclosable credentials.
         */
        public ResultOfOperation computeAccessDecision()
        {
              ResultOfOperation resComputeAccessDecision = new ResultOfOperation();

              // check if the client has enough access rights to get the requested service
              ResultOfOperation resDeduction = doDeduction();

              if(resDeduction.isSuccessful())
              {
                  // if the check succeeds then return the result to the TN Layer
                  if( ((Integer) resDeduction.getOperationResult()).compareTo(new Integer(1)) == 0)
                  {
                        deleteTempFiles();
                        return resDeduction;
                  }
                  // otherwise compute the sets of abducibles
                  ResultOfOperation rescomAbducibles = computeAbducibles();
                    
                  if(rescomAbducibles.isSuccessful())
                  {
                      // in case there is no disclosable set then deny access
                      if(rescomAbducibles.getOperationResult() == null)
                      {
                               deleteTempFiles();
                               resComputeAccessDecision.setIsSuccessful(true);
                               resComputeAccessDecision.setOperationResult(new Integer(0));
                               return resComputeAccessDecision;
                      }
                      // if disclosable set then compute the minimal set of missing credentials 
                      ResultOfOperation resAbduction = doAbduction((LinkedList) rescomAbducibles.getOperationResult());
                          
                      if(resAbduction.isSuccessful())
                      {
                          // if no missing set found then deny
                          if(resAbduction.getOperationResult() == null)
                          {
                                    deleteTempFiles();
                                    resComputeAccessDecision.setIsSuccessful(true);
                                    resComputeAccessDecision.setOperationResult(new Integer(0));
                                    return resComputeAccessDecision;
                          }
                          else // if there is a missing set CM then return it.
                          {
                                  deleteTempFiles();
                                  return resAbduction;
                          }
                          
                      }
                      else
                      {
                          System.err.println("IACCESS: EXCEPTION: BasicAccessControlAlgorithm.doAbduction(): " + resAbduction.getOperationResult().toString());
                          deleteTempFiles();
                          return resAbduction;
                      }
                  }
                  else
                  {
                      System.err.println("IACCESS: EXCEPTION: BasicAccessControlAlgorithm.computeAbducibles(): " + rescomAbducibles.getOperationResult().toString());
                      deleteTempFiles();
                      return rescomAbducibles;
                  }
              }
              else
              {
                  System.err.println("IACCESS: EXCEPTION: BasicAccessControlAlgorithm.doDeduction(): " + resDeduction.getOperationResult().toString());
                  deleteTempFiles();
                  return resDeduction;
              }
        }
}
