/*
 * Copyright 2006, 2008 Hristo Koshutanski, Aliaksandr Lazouski, Associazione CREATE-NET, and Universita' degli Studi di Trento.  
 * 
 * This file is part of iAccess authorization service software (called iAccess).
 * iAccess is free software: you can redistribute it and/or modify 
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * iAccess is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with iAccess.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.iaccess.TNProtocol;

import org.iaccess.CredentialManager.*;

import java.io.File;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.StringReader;
import java.util.regex.Pattern;
import java.util.logging.Level;
import java.util.Calendar;
import org.xml.sax.InputSource;
import javax.xml.parsers.DocumentBuilderFactory;
import org.w3c.dom.Document;
import java.util.Vector;

/**
 * Keeps all (important) session information for a negotiation process.
 * It also keeps relevant information loaded from the negotiation level configuration.
 * @see org.iaccess.TNProtocol.TRequest
 */
public class TSessionInfo{
        /** In milliseconds. 40 seconds is the default value. */
	public static long SessionTimeLimit = 40*1000;
        /** In milliseconds. Takes the SessionTimeLimit by default.*/
	public static long DynamicSessionTimeExtension = SessionTimeLimit;
        /** In milliseconds. The validity of the (SAML) Authorization statement.*/
        public static long AccessDecisionTimeValidity = SessionTimeLimit;
        /** Defines the port of JBOSS TCP/IP Server for incoming messages. By default is 5555.
         * Use different ports for incoming and outcoming messages if you are running TN Server and TN Client on one machine. */
	public static int TCPportIn = 5555;
        /** Defines the port of JBOSS TCP/IP Server for outcoming messages. By default is 5555
         * Use different ports for incoming and outcoming messages if you are running TN Server and TN Client on one machine. */
	public static int Default_TCPportOut = 5555;
	/**Updated dynamically from the incoming requests. */
    public int TCPportOut;
        /** Defines the IP address of the remote iAccess server. */
        public static String IPaddress = "localhost";
	/** Specifies MinimalityCriterion (ON/OFF).*/
	public static String MinimalityCriterion = new String();
        /** Specifies full path to the config_dlv.xml file.*/
	public static String DLVWrapperConfigurationFile = new String();
	/** Specifies the root folder of the system. */
	public static String RootDir = new String();
        /** Specifies the system variable name that is instantiated in any configuration file. */
        public static String ConfigRootDir = "IACCESS_ROOT";
        /** Specifies the platofrm dependant file separator '\' or '/'.
         Default separator is the unix one '/'. */
        public static String ActualFileSeparator = "/";
        /** Specifies OS type. 1-Linux; 2-Windows; 3-MacOSX PPC; 4-MacOSX Intel
         * Default one is the Linux type */
        public static int OSType = 1;
        /** Keeps the distinguished identity of a SessionInfo object.
         * It is used to update CP and CN as a hashtable key.
         * It is extratced from X.500 field of the X.509 certificate acompaning the private key used for the SSL channel.
         */
        public String Identity = null;
        /** Stores opponent's CredAnyValurTerm. Initially, CredAnyValurTermOpponent = CredAnyValurTerm .*/
	private String CredAnyValueTermOpponent = null;
        /** Keeps information about all theads serving the negotiation process.*/
	public ThreadGroup TNThreadGroup;	
        /** Keeps the time when the SessionInfo is created, i.e. the initial
         *time when the negotiation started. In milliseconds. Only for time mesuarment.*/
        public long negStartTime; 
        /** Keeps information about the main (initial) service request. */
        public String ServiceRequest = new String();
        /**Indicates if active credentials are to be shared in multi-session negotiations between same entities.*/
        public static boolean MultiSessionSharedCreds = false;
        /** A counter of the threads, which have already created but not started the Trust Negotiation Protocol. */
	private int BlockCnegCounter = 0;
	/** Specifies the time when a session expires.*/
	private long ValidNotAfter;
        
        /** Defines a specific term used to identify any value in a field of a credential request
         * Initialized in TSessionInfo. Experimental field. Do not use if not sure why. */
        public static String CredAnyValueTerm = null ;
        
        /** Defines a specific mode for authorization handling.
         * If false then the classical mode of intiating negotiation-based authorization with an opponent agent.
         * If true then a server-based mode where the agent is awaiting on a specific call from a PEP to get an access
         * decision. */
        public static boolean IsServerAuthorizationMode = false;

        /** Private objects for presented and declined credentials/certificates. */
	private TCN cn_Credentials = new TCN();
	private TCP cp_Credentials = new TCP();	
	
        /** Creates a new instance of TSessionInfo.
         * @param Request Must be invoked as a request for a service. It returns immediately if not. */
	public TSessionInfo(TRequest Request) {
              try{  
                if (Request.IsService != true)
                    return;
                ServiceRequest = new String(Request.ServiceRequest);
		TNThreadGroup = new ThreadGroup(Request.SessionID);
                negStartTime = Calendar.getInstance().getTimeInMillis();
		ValidNotAfter = negStartTime + SessionTimeLimit;	
                Identity = new String(Request.Identity);
                if( (Request.Credentials!=null) && (Request.Credentials.length>0) && (Request.Credentials[0].length()>0) ) 
                    CredAnyValueTermOpponent = new String(Request.Credentials[0]);
                else
                    CredAnyValueTermOpponent = CredAnyValueTerm;

                if( (Request.Credentials!=null) && (Request.Credentials.length>1) ) 
                {
                    TCPportOut = Integer.parseInt(Request.Credentials[1]);
                }
                else
                    TCPportOut = Default_TCPportOut;
                //set session start time
                
              }catch(Exception e)
              {
                  System.err.println("IACCESS: EXCEPTION: TSessionInfo(TRequest): "+e);
              }
        
        }
	
        /** Internal class defining the set of declined credentials/certificates. */
	private class TCN {		
		private Vector CN = new Vector();
		
		synchronized void UpdateCN(String NewCN) {
			if ( (NewCN != null) && (!CN.contains(NewCN)) )
				CN.addElement(NewCN);			
		}	
		
		synchronized Vector GetCN(){		
			return (Vector) CN.clone();
		}
	}
        
        /** Internal class defining the set of presented credentials/certificates. */
	private class TCP {		
		private Vector CP = new Vector();
		
		synchronized void UpdateCP(String NewCP) {
			if ( (NewCP != null) && (!CP.contains(NewCP)) )
                        {
                            CP.addElement(NewCP);	
                            //USER PROFILE UPDATE Use of synchronized method
                            UserProfile.updateOngoingProfile( NewCP.substring(0, NewCP.indexOf(",")), NewCP);
                        }
		}
		
		synchronized void UpdateCP(String[] NewCP) {
			if (NewCP != null)
				for (int i = 0; i < NewCP.length; i++) 
					if ( (NewCP[i] != null) && (!CP.contains(NewCP[i])) )
                                        {
						CP.addElement(NewCP[i]);			
                                                //USER PROFILE UPDATE Use of synchronized method
                                                UserProfile.updateOngoingProfile( NewCP[i].substring(0, NewCP[i].indexOf(",")), NewCP[i]);
                                        }
		}
		

                
		synchronized void UpdateCP(Vector NewCP) {
			if (NewCP != null)
				for (int i = 0; i < NewCP.size(); i++) 
					if ( (NewCP.get(i) != null) && (!CP.contains(NewCP.get(i))) )
                                        {
						CP.addElement(NewCP.get(i));			
                                                //USER PROFILE UPDATE Use of synchronized method
                                                UserProfile.updateOngoingProfile( ((String)NewCP.get(i)).substring(0, ((String)NewCP.get(i)).indexOf(",")), (String)NewCP.get(i));
                                        }
		}
                
                
                synchronized Vector GetCP(){
			return (Vector) CP.clone();
		}
	}	
	
        /** Adds a new declined credential/certificate to the existing set of declined credentials.
         * It call a synchronized method within the class which defines the set of declined credentials/certificates.
         * @see org.iaccess.TNProtocol.TSessionInfo.TCN
         */
	public void UpdateCN(String NewCN) {
		cn_Credentials.UpdateCN(NewCN);
	}
        
	/** Adds a new presented credential/certificate to the existing set of presented credentials.
     * It calls a synchronized method within the class which defines the set of presented credentials/certificates.
     * @see org.iaccess.TNProtocol.TSessionInfo.TCP
     */
	public void UpdateCP(String NewCP) {
		cp_Credentials.UpdateCP(NewCP);
	}
        
	/** Adds a set of presented credentials/certificates to the existing set of presented credentials.
     * It calls a synchronized method within the class which defines the set of presented credentials/certificates.
     * @see org.iaccess.TNProtocol.TSessionInfo.TCP
     */
	public void UpdateCP(String[] NewCP) {
		cp_Credentials.UpdateCP(NewCP);
	}

	/** Adds a set of presented credentials/certificates to the existing set of presented credentials.
     * It calls a synchronized method within the class which defines the set of presented credentials/certificates.
     * @see org.iaccess.TNProtocol.TSessionInfo.TCP
     */
	public void UpdateCP(Vector NewCP) {
		cp_Credentials.UpdateCP(NewCP);
	}
        
    /** Returns the current set of declined credentials/certificates.
     * It calls a synchronized method within the class which defines the set of declined credentials/certificates.
     * @see org.iaccess.TNProtocol.TSessionInfo.TCN
     */
	public Vector GetCN(){
		return cn_Credentials.GetCN();
	}
        
	/** Returns the current set of presented credentials/certificates.
     * It calls a synchronized method within the class which defines the set of presented credentials/certificates.
     * @see org.iaccess.TNProtocol.TSessionInfo.TCP
     */
	public Vector GetCP(){
		return cp_Credentials.GetCP();
	}	
        
	/** Returns the time in milliseconds when a session expires. */
	public synchronized long GetValidNotAfter() {
		return ValidNotAfter;
	}
                

    /** Sets the SetCredAnyValueTermOpponent. */
	public synchronized void SetCredAnyValueTermOpponent(String CredAnyValueTermOpponent) {
		this.CredAnyValueTermOpponent = CredAnyValueTermOpponent;
	}	
        

    /** Gets the SetCredAnyValueTermOpponent. */
	public synchronized String GetCredAnyValueTermOpponent() {
		return CredAnyValueTermOpponent;
	}	

        
    /** Sets the time in milliseconds when session should expire*/
	public synchronized void SetValidNotAfter(long ValidNotAfter) {
		this.ValidNotAfter = ValidNotAfter;
	}	
                
        
    /** Increases the counter of the threads already been created but not started the trust negotiation protocol. */
	public synchronized void IncreaseBlockCnegCounter() {
		BlockCnegCounter++;
	}
        
	/** Decreases the counter when a new thread starts the trust negotiation protocol,
     * or awaits for the result on the existing thread of same request. */
	public synchronized void DecreaseBlockCnegCounter() {
		BlockCnegCounter--;
		notifyAll();
	}
        
        /** Removes current thread from the negotiation process,
         * i.e. every new thread with a same request will start a new Trust Negotiation Protocol.
         * Each thread has 3 time states:
         * <ul>
         * <li> initializing (name of the thread in this case usually starts with "TN: ");
         * <li> in negotiations (name of the thread in this case usually starts with "*TN: ");
         * <li> destroying (name of the thread in this case usually starts with "***TN: ").
         * </ul>
         * So, we simply change the name of a current thread, what indicates its new state.
         * Also we must asshure that there is no any initializing thread. */
	public synchronized void RemoveCneg() {
		try {
			while (true) {
				if (BlockCnegCounter != 0) {
					wait();
				} else {
					Thread.currentThread().setName("***" + Thread.currentThread().getName()); 
					return;
				}
			}	
		} catch (Exception e) {
			iAccessDispatcher.logger.logp(Level.SEVERE, "TSessionInfo", "RemoveCneg", "Thread: "+ Thread.currentThread().getName() +"; EXCEPTION: "+ e.toString());
		}
	}
        
	/** Loads session information from config.xml file.
     * @exception Exception File path is not valid, or format of Session Configuration info is not correct, or unrecognized OS type.
     */
	public static synchronized void LoadSessionConfig(String ConfigPath) throws Exception {		
            
            if(Pattern.compile("\\s").matcher(System.getProperty("user.dir")).find())
            {
                //System.out.println("Root dir: " + System.getProperty("user.dir"));
                throw new Exception("Root path contains blank spaces. Run iAccess with no blank spaces in the root path.");
            }
            
            // Checking the config file path
            File configFile = new File(ConfigPath);
            if(configFile.isFile())
                ConfigPath = configFile.getPath();
            else
                throw new Exception("The specified configuration file does not exist!");
                    

            String os_name = System.getProperty("os.name");
            String os_arch = System.getProperty("os.arch");
            if(os_name.startsWith("Mac OS X"))
                OSType = 3;
            else
                if(os_name.startsWith("Windows"))
                    OSType = 2;
                else
                    if(os_name.startsWith("Linux"))
                        OSType = 1;
                    else
                        throw new Exception("Unsupported Operating System Type: " + os_name + " with " + os_arch);
            

            BufferedReader configFileToModify = new BufferedReader( new FileReader(ConfigPath) );
            String readLine,
                   newConfigFile = "";
            RootDir = System.getProperty("user.dir");
            String myrootDir = new String(RootDir);
            while((readLine = configFileToModify.readLine()) != null)
                newConfigFile += readLine;
            
            //Making OS dependant file separators.
            if(OSType==2)
            {
                //Because root replacement part below removes C:\file\ to C:file. 
                myrootDir = myrootDir.replace('\\',';');   
                newConfigFile = Pattern.compile(ConfigRootDir).matcher(newConfigFile).replaceAll(myrootDir);
                newConfigFile = newConfigFile.replace(';','\\');
                ActualFileSeparator = "\\"; 
            }
            else
            {
                newConfigFile = Pattern.compile(ConfigRootDir).matcher(newConfigFile).replaceAll(myrootDir);
                newConfigFile = newConfigFile.replace(';','/');
                ActualFileSeparator = "/";
            }

            //Instantiating the root dir variable in the config file           
            // Create a builder factory
            DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
            factory.setValidating(false);
            // Create the builder and parse the file
            Document doc = factory.newDocumentBuilder().parse(new InputSource( new StringReader(newConfigFile)));
	    SessionTimeLimit = 1000* Long.parseLong(doc.getDocumentElement().getElementsByTagName("SessionTimeLimit").item(0).getFirstChild().getNodeValue().trim());
	    DynamicSessionTimeExtension = 1000 * Long.parseLong(doc.getDocumentElement().getElementsByTagName("DynamicSessionTimeExtension").item(0).getFirstChild().getNodeValue().trim());
            AccessDecisionTimeValidity = 1000 * Long.parseLong(doc.getDocumentElement().getElementsByTagName("AccessDecisionTimeValidity").item(0).getFirstChild().getNodeValue().trim());
            
	    TCPportIn  = Integer.parseInt(doc.getDocumentElement().getElementsByTagName("TCPportIn").item(0).getFirstChild().getNodeValue().trim());
	    Default_TCPportOut = Integer.parseInt(doc.getDocumentElement().getElementsByTagName("TCPportOut").item(0).getFirstChild().getNodeValue().trim());
	    
            MinimalityCriterion = doc.getDocumentElement().getElementsByTagName("MinimalityCriterion").item(0).getFirstChild().getNodeValue().trim();
	    DLVWrapperConfigurationFile = doc.getDocumentElement().getElementsByTagName("DLVConfigFilePath").item(0).getFirstChild().getNodeValue().trim();
            
            CredentialsConfig.DEFAULT_KEYS = doc.getDocumentElement().getElementsByTagName("SSLkeystore").item(0).getFirstChild().getNodeValue().trim();
            CredentialsConfig.DEFAULT_PASSWORD = doc.getDocumentElement().getElementsByTagName("TrustedStorePassword").item(0).getFirstChild().getNodeValue().trim();
            CredentialsConfig.DEFAULT_TRUSTSTORE = doc.getDocumentElement().getElementsByTagName("TrustedCAStore").item(0).getFirstChild().getNodeValue().trim();            
            
            
            IPaddress = doc.getDocumentElement().getElementsByTagName("IPaddressByName").item(0).getFirstChild().getNodeValue().trim();
            
            CredentialsConfig.CaKeys = doc.getDocumentElement().getElementsByTagName("TrustedSOAStore").item(0).getFirstChild().getNodeValue().trim();
            CredentialsConfig.CaKeysPassword = doc.getDocumentElement().getElementsByTagName("TrustedStorePassword").item(0).getFirstChild().getNodeValue().trim();
            CredentialsConfig.CertificateDB = doc.getDocumentElement().getElementsByTagName("CertificatesDB").item(0).getFirstChild().getNodeValue().trim();
            
            String multisessioncred = doc.getDocumentElement().getElementsByTagName("MultiSessionSharedCreds").item(0).getFirstChild().getNodeValue().trim();
            if(multisessioncred.equals("ON")) MultiSessionSharedCreds = true;
            
            String needClientAuth = doc.getDocumentElement().getElementsByTagName("SSLmutualauthentication").item(0).getFirstChild().getNodeValue().trim();        
            if ( needClientAuth.equals("ON")) CredentialsConfig.needClientAuth = true;
            
            
            String ifServerAuthzMode = doc.getDocumentElement().getElementsByTagName("IsServerAuthorizationMode").item(0).getFirstChild().getNodeValue().trim();        
            if ( ifServerAuthzMode.equals("ON")) IsServerAuthorizationMode = true;
            
            CredAnyValueTerm = doc.getDocumentElement().getElementsByTagName("CredAnyValueTerm").item(0).getFirstChild().getNodeValue().trim();
            Character c = new Character(CredAnyValueTerm.charAt(0));
            Character c_a = new Character('a');
            Character c_z = new Character('z');
            if( c.compareTo(c_a) < 0 && c.compareTo(c_z)>0 )
			throw new Exception("Incorrect Values For CredAnyValueTerm. It has to start with a letter in the range 'a'..'z'. ");            

            if(SessionTimeLimit <0 || DynamicSessionTimeExtension <0 || AccessDecisionTimeValidity <0)
			throw new Exception("Incorrect Values For SessionTimeLimit and/or DynamicSessionTimeExtension and/or AccessDecisionTimeValidity");

            if(Default_TCPportOut <=0 || Default_TCPportOut <=0)
			throw new Exception("Incorrect Values For TCPportIn and/or TCPportOut");


            String holder = CertificateManager.loadSSLKeystoreAndGetHolderDNToLogic();
            if(holder!=null)
                    CredentialsConfig.principalName = holder;
               else
                    throw new Exception("Unable to retrieve holder element of an X509 Certificate in the SSL keystore.");            
            
	}
}