/*
 * Copyright 2006, 2008 Hristo Koshutanski, Aliaksandr Lazouski, Associazione CREATE-NET, and Universita' degli Studi di Trento.  
 * 
 * This file is part of iAccess authorization service software (called iAccess).
 * iAccess is free software: you can redistribute it and/or modify 
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * iAccess is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with iAccess.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.iaccess.TNProtocol;

import java.util.*;
import java.util.logging.*;
import java.net.*;
import java.io.*;
import javax.net.ssl.SSLSocket;

import org.iaccess.CredentialManager.*;


/**
 * Manages all negotiation processes on a hight level. It dispatches them according to a session information. It only forwards requests in a proper way.
 * It extends the abstract class java.lang.Thread, threfore always starts in a new thread. 
 * Once the iAccessDispatcher is run, it awaits for service requests or credential requests.
 * The class TNProtocol.TRequest defines the type/structure of an incoming request. 
 * For each received request a new object of TNProtocol.iAccessNegotiation class is created and run (always started) in a new thread.
 * For each service request a new object of the TSessionInfo class is generated. 
 * <p>This class supports multiple sessions going concurrently.
 * <p>More details are given to each of the functions in the class.
 * @see org.iaccess.TNProtocol.TRequest
 * @see org.iaccess.TNProtocol.iAccessNegotiation
 * @see org.iaccess.TNProtocol.TSessionInfo
 * @see org.iaccess.TNProtocol.JBOSS
 * @see org.iaccess.CredentialManager.CertificateManager
 */
public class iAccessDispatcher extends Thread{
            
        public static int inCredentials = 0;
        public static int outCredentials = 0;
    
        /** Indicates the current object's state.
         * stopWorking is true means that object is on the way to destruction.
         * By default stopWorking = true*/
	private volatile boolean stopWorking = true;	
        /** Indicates the current object's state.
         * notReady is true means that iAccessDispacher is processing the request and has been blocked for new requests.
         * By default notReady = false*/
	private volatile boolean notReady = false;
	/** Specifies the current processing request or the last processed request. */
	public TRequest Request;	
        /** Specifies the addressee of the request.
         * TCP/IP sockets are used. This variable should be replaced in the next version by more complete value. */
	private SSLSocket SocketWaitingResult;
        /** Keeps the links to the SessionInfo for all active sessions going concurrently. */
	public static Hashtable SessionInfoTable = new Hashtable();
        /** Specifies the output stream which gives the description of all events occured during the negotiation processes.
         * Actually, this information is forwarded to the application which has created the instance of iAccessDispatcher. 
         * If you don't need some additional comments put <b>null</b> here. */
	public static OutputStream OutPutStream = null;
	public static long UnexpectedErrorSleepTime = 30;
        /** Indicates the number of received requests. It allows to distinguish requests within one session. */
	public static long InteractionCounter = 0;
	
        // Log valiables
	public static Logger logger = Logger.getLogger("iAccessDispatcherLogger");
	private FileHandler LoggerHandler;
	private ErrorManager LogErrManager = new ErrorManager();

        /** Creates and starts a new instance of the iAccessDispatcher. Class java.util.Thread is inherited by iAccessDispatcher,
         * i.e. iAccessDispatcher always runs in a new thread.
        * @param ConfigFilePath specifies full path to the SessionConfig.xml file. Set to null if SessionConfig Info has been already loaded.
        * @param LogFilePath    specifies full path to the logger file. If is null DefaultiAccessTNProtocol.log file will be created in the current working directory.
        * @param OutPutStream   specifies the output stream which returns the description of all events occured during the negotiation processes to the calling entity. 
        */
	public iAccessDispatcher(String ConfigFilePath, String LogFilePath, OutputStream OutPutStream) {
		super("iAccessDispatcher");
                System.out.println("\n*************************************************************************");
                System.out.println("iAccess Copyright (C) 2006, 2008 Hristo Koshutanski, Aliaksandr Lazouski, ");
                System.out.println("Associazione CREATE-NET, and Universita' degli Studi di Trento.");
                System.out.println("This program comes with ABSOLUTELY NO WARRANTY. This is free software,");
                System.out.println("and you are welcome to redistribute it under certain conditions.");
                System.out.println("For details look at the GNU Lesser General Public License received alogn ");
                System.out.println("with iAccess. If not, see <http://www.gnu.org/licenses/>.");
                System.out.println("*************************************************************************");
                System.out.println("********************* IACCESS AUTHORIZATION SERVICE *********************");
		iAccessDispatcher.OutPutStream = OutPutStream;
                // Initialize Logger		
		try {
			LoggerHandler = new FileHandler(LogFilePath,true);
                       
		} catch (Exception e){
			try {				
				LoggerHandler = new FileHandler("DefaultiAccessTNProtocol.log",true);
			} catch (Exception err){
				System.err.println("IACCESS: EXCEPTION: iAccessDispatcher initialization: " + err.toString());
			}
		}
		LoggerHandler.setErrorManager(LogErrManager);
		iAccessDispatcher.logger.addHandler(LoggerHandler);
		
//		Load Session's Configuration and Users Profiles
		if (ConfigFilePath == null)
                {
                    iAccessDispatcher.logger.logp(Level.WARNING, "iAccessDispatcher", "constructor", "Cannot load config data. Make sure that config.xml is in: "+ System.getProperty("user.dir") +"\nIACCESS: iAccessDispatcher not loaded!");
                }
                else
                {
			try {   // Loads the main config.xml file
				TSessionInfo.LoadSessionConfig(ConfigFilePath);
                                System.out.println("IACCESS: Configuration initialized.");
                                
                                //Load in memory the trusted keys of CAs and SOAs.
                                CertificateManager.loadTrustedPublicKeys();

                                System.out.println("IACCESS: Initializing own credentials...");	
                                CertificateManager.loadMyCredentials();
                                System.out.println("IACCESS: Own credentials initialized.");	

                                //iAccessDispatcher.logger.logp(Level.INFO, "iAccessDispatcher", "constructor", "iAccessDispatcher is initialized");
                                //Runs the dispatcher.
                                start();

            } catch(Exception e){
				iAccessDispatcher.logger.logp(Level.SEVERE, "iAccessDispatcher", "constructor", "Cannot load config data. Make sure that config.xml is in: "+ System.getProperty("user.dir") + "\nIACCESS: EXCEPTION: " + e.toString()+"\nIACCESS: iAccessDispatcher not loaded!");
			}		
                }
		//start();
	}

        /** Processes a new recevied request for a service.
        * For each request it runs a new Trust Negotiation Protocol (creates an object of TNProtocol.iAccessNegotiation class)
        * which always starts in a new thread.
        * For each request a new object of TSessionInfo is generated.
        * Returns the set of opponent's active credentials out of the negotiation process.
        * @see org.iaccess.TNProtocol.iAccessNegotiation
        */
	private void startNewSessionServer(TRequest Request, SSLSocket SocketWaitingResult){ //Asynchronous
		try {
                        //If SessionInfo does not exist throw an exception
                        //The TN protocol requires incoming service requests to specify
                        //a unique sessionID for authorization/negotiaiton reference.
                        //Here, one can generate a new sessionID, but in a future version of the protocol.
                        if (Request.SessionID == null ) {
                            System.err.println("IACCESS: ERROR: Received Service Request with no SessionID. TNProtocol requires such.");
                            throw new Exception("Received Service Request with no SessionID. TNProtocol requires such.");
			            }
                        else
                        {  //Check correctness of SessionID generation, i.e. it should not exist (must be unique).
                            if(iAccessDispatcher.SessionInfoTable.containsKey(Request.SessionID)){
                                System.err.println("IACCESS: ERROR: Received Service Request with a dublicated SessionID. TNProtocol requires to be unique.");
                                throw new Exception("Received Service Request with a dublicated SessionID. TNProtocol requires to be unique.");
                            }
                        }
                        //Check correctness of the declared Identity ID in the service request
                        //TO BE CHECKED IF THE IDNETITY IN A REQUEST IS THE ONE 
                        //OF THE SSL CERTIFICATE IN THE HANDSHAKE!
                        //TO BE DONE IN A NEXT IACCESS RELEASE!

                        TSessionInfo SessionInfo = new TSessionInfo(Request);
                        //Update CP with exsting credentials from the opponent's profile
                        if(TSessionInfo.MultiSessionSharedCreds)
                           SessionInfo.UpdateCP(UserProfile.getCredentials(Request.Identity, SessionInfo.GetValidNotAfter()+TSessionInfo.DynamicSessionTimeExtension));
                        else
                           SessionInfo.UpdateCP(UserProfile.getEntityLogicIDCredOfSSL(Request.Identity));
                            
                        iAccessDispatcher.SessionInfoTable.put(Request.SessionID, SessionInfo);

                        if (iAccessDispatcher.OutPutStream != null)
                            iAccessDispatcher.OutPutStream.write( ("*** NEW SESSION <" + Request.SessionID+"> ***"+"}"+Request.SessionID).getBytes());

                        //Run negotiation on the received Request
                        new iAccessNegotiation(Request, SocketWaitingResult).runNegotiations();

		} catch (Exception e) {
			iAccessDispatcher.logger.logp(Level.SEVERE, "iAccessDispatcher", "startNewSessionServer", "Thread: "+ Thread.currentThread().getName() +"; EXCEPTION: "+ e.toString());
			try {
				JBOSS.sendResponseSSL(Request, iAccessNegotiation.deny, SocketWaitingResult);
			} catch (Exception ee){iAccessDispatcher.logger.logp(Level.SEVERE, "iAccessDispatcher", "startNewSessionServer", "Thread: "+ Thread.currentThread().getName() +"; EXCEPTION: "+ e.toString());}
        }
	}

        /** Processes a new service request which has to be sent to an opponent's trust negotiation agent.
        * This is a synchronous method. It calls TNProtocol.JBOSS.sendRequest(...) method to send request to the opponent's agent and awaits for a result.
        * For each request a new object of TSessionInfo is generated. 
        * @see org.iaccess.TNProtocol.JBOSS
        * @see org.iaccess.UserInterface.UserInterface
        */
	public String startNewSessionClient(String ServiceName, String ServiceAction,long SessionTimeLimit, long TimeExtemsion, InetAddress ServerInetAddress, int ServerPort){ //Synchronous

    		    //Set up the new session info
                TSessionInfo.SessionTimeLimit = SessionTimeLimit * 1000; 
    		    TSessionInfo.DynamicSessionTimeExtension = TimeExtemsion * 1000;

                String ServiceRequest = CertificateLogical.makeLogical(ServiceName.toLowerCase())+","+CertificateLogical.makeLogical(ServiceAction.toLowerCase());
                //Generate a tructure for a protocol request. Specify true for IsService field.
                //Generate SessionID
                Calendar calendar = Calendar.getInstance();
                String SessionID = Long.toString(calendar.getTimeInMillis()) + Double.toString(Math.random());
                TRequest Request = new TRequest(SessionID, true, ServiceRequest, CredentialsConfig.principalName, TSessionInfo.CredAnyValueTerm, TSessionInfo.TCPportIn);
                TSessionInfo SessionInfo = new TSessionInfo(Request);
                SessionInfo.TCPportOut = ServerPort;
                //Set up the new session info object in the dispatcher hashtable
                iAccessDispatcher.SessionInfoTable.put(Request.SessionID, SessionInfo);			
                String newResult = null;
                int Result = iAccessNegotiation.notReady;

                try {
                       //Send the request to an opponent's agent (by IP) and await for the response
			while(stopWorking)
				Thread.sleep(UnexpectedErrorSleepTime);
			if (iAccessDispatcher.OutPutStream != null)
				iAccessDispatcher.OutPutStream.write( (">>>"+Request.ServiceRequest+"!outgoing service request: submitted" + "}" + Request.SessionID).getBytes());
                        
            //result is null if deny, otherwise a SAML token is received with a permit decision.
            newResult = JBOSS.sendRequestSSL(Request, ServerInetAddress);
            if ( newResult == null)
                Result = iAccessNegotiation.deny;
            else
                Result = iAccessNegotiation.grant;

                } catch (Exception e) {			
			iAccessDispatcher.logger.logp(Level.SEVERE, "iAccessDispatcher", "startNewSessionClient", "Thread: "+ Thread.currentThread().getName() +"; EXCEPTION: "+ e.toString());
			Result = iAccessNegotiation.deny;
		} finally {
                       //Await all other threads in session and remove session info
			try {
				if (iAccessDispatcher.OutPutStream != null){
					String TempString = "notReady";
					if (Result == iAccessNegotiation.deny) TempString = "denied";
					if (Result == iAccessNegotiation.grant) TempString = "granted";
                                        String negTime = String.valueOf(Calendar.getInstance().getTimeInMillis()-((TSessionInfo)this.SessionInfoTable.get(Request.SessionID)).negStartTime);
					iAccessDispatcher.OutPutStream.write( ("<<<"+ Request.ServiceRequest + "!" + "incoming service response: "+TempString + " ("+negTime+" ms)" + "}" + Request.SessionID).getBytes());
				}

				Thread CurrSessionThreads[] = new Thread[SessionInfo.TNThreadGroup.activeCount()];
				SessionInfo.TNThreadGroup.enumerate(CurrSessionThreads);
				for (int i = 0; i < CurrSessionThreads.length; i++) { 
					if ( (!CurrSessionThreads[i].equals(Thread.currentThread())) && CurrSessionThreads[i].isAlive())
						CurrSessionThreads[i].join();
				}
			} catch (Exception e) {
				iAccessDispatcher.logger.logp(Level.SEVERE, "iAccessDispatcher", "startNewSessionClient", "Thread: "+ Thread.currentThread().getName() +"; EXCEPTION: "+ e.toString());
			}
                        System.out.println("IACCESS: Negotiation Time (ms): " + String.valueOf(Calendar.getInstance().getTimeInMillis()-((TSessionInfo)this.SessionInfoTable.get(Request.SessionID)).negStartTime));
			iAccessDispatcher.SessionInfoTable.remove(Request.SessionID);
                        
		}
                //Store the access decision in the "log/" folder. 
                try{
                if(newResult!=null)
                {
                    FileOutputStream fout = new FileOutputStream(new File(System.getProperty("user.dir")+TSessionInfo.ActualFileSeparator+"log"+TSessionInfo.ActualFileSeparator+Request.SessionID+"_SAMLToken.xml"));
                    fout.write(newResult.getBytes());
                    fout.close();
                }
                }catch(Exception e)
                {
                    System.err.println("IACCESS: EXCEPTION: Unable to store the SAML decision token in IACCESS_ROOT/log folder: " + e);
                }
                return newResult;
	}

        /** Processes a new credential request recevied.
        * For each request it runs a new trust negotiation protocol (creates an object of TNProtocol.iAccessNegotiation class)
        * which always starts in a new thread.
        * For each request an existing object of TSessionInfo is updated. 
        * @see org.iaccess.TNProtocol.iAccessNegotiation
        */        
	private void joinExistSession() {
		try {
                        //If Session Time has expired return deny
			if ( ! iAccessDispatcher.SessionInfoTable.containsKey(Request.SessionID) )
                        {
                            JBOSS.sendResponseSSL(this.Request, iAccessNegotiation.deny, SocketWaitingResult);
                            return;
                        }
                        if( iAccessDispatcher.SessionInfoTable.get(Request.SessionID) != null && ((TSessionInfo) iAccessDispatcher.SessionInfoTable.get(Request.SessionID)).GetValidNotAfter() <= System.currentTimeMillis()) 
                        {
                            JBOSS.sendResponseSSL(this.Request, iAccessNegotiation.deny, SocketWaitingResult);
                            return;
			}
                        //Already existing session. 
                        //Set CredAnyValueTermOpponent in the session info.
                        if( (Request.Credentials!=null) && (Request.Credentials.length>0) && (Request.Credentials[0].length()>0))
                            ((TSessionInfo) iAccessDispatcher.SessionInfoTable.get(Request.SessionID)).SetCredAnyValueTermOpponent(Request.Credentials[0]);
                        
                        //Check for a special anyvalue term in a holder field
                        String matchCred = CertificateManager.getCredential(Request.ServiceRequest, TSessionInfo.CredAnyValueTerm);
                        if(matchCred!=null)
                        {
                            //Write to log/stdoutput request what has been requested 
                            if(Request.ServiceRequest.startsWith(TSessionInfo.CredAnyValueTerm))
                                System.out.println("IACCESS: Substituted Request: "+Request.ServiceRequest+" with: "+matchCred);
                            Request.ServiceRequest = matchCred;
                        }
                        //Updating CP if new credentials have been presented meanwhile
                        if( TSessionInfo.MultiSessionSharedCreds )
                        {
                            long ln1 = ((TSessionInfo) iAccessDispatcher.SessionInfoTable.get(Request.SessionID)).GetValidNotAfter();
                            long ln2 = ((TSessionInfo) iAccessDispatcher.SessionInfoTable.get(Request.SessionID)).negStartTime;
                            if(ln2+TSessionInfo.SessionTimeLimit < ln1)
                               ((TSessionInfo) iAccessDispatcher.SessionInfoTable.get(Request.SessionID)).UpdateCP(UserProfile.getCredentials(Request.Identity, ln1));
                            else
                               ((TSessionInfo) iAccessDispatcher.SessionInfoTable.get(Request.SessionID)).UpdateCP(UserProfile.getCredentials(Request.Identity, ln1+TSessionInfo.DynamicSessionTimeExtension));
                        }
                        else //update entity ID from the SSL session. 
                           ((TSessionInfo) iAccessDispatcher.SessionInfoTable.get(Request.SessionID)).UpdateCP(UserProfile.getEntityLogicIDCredOfSSL(Request.Identity));
                        
                        //Run the iAccessNegotiation protocol 
			new iAccessNegotiation(Request, SocketWaitingResult).runNegotiations();
		} catch (Exception e) {
			iAccessDispatcher.logger.logp(Level.SEVERE, "iAccessDispatcher", "joinExistSession", "Thread: "+ Thread.currentThread().getName() +"; Error: "+ e.toString());
			try {
				JBOSS.sendResponseSSL(this.Request, iAccessNegotiation.deny, SocketWaitingResult);
			} catch (Exception ee){iAccessDispatcher.logger.logp(Level.SEVERE, "iAccessDispatcher", "joinNewSessionServer", "Thread: "+ Thread.currentThread().getName() +"; EXCEPTION: "+ ee.toString());}
		}
	} 

        /** Notifies iAccessDispatcher about newly received requests and initiates iAccessDispatcher processing.
        * This method is called by a lower level server, as from the JBOSS class in our case.
        * @param Request the received request description.
        * @param SocketWaitingResult specifies the address of the request.
        * @see org.iaccess.TNProtocol.JBOSS
        */                
	public synchronized void receiveRequest(TRequest Request, SSLSocket SocketWaitingResult) {
		if (notReady) 
			try {
				wait();
			} catch (Exception e) {iAccessDispatcher.logger.logp(Level.SEVERE, "iAccessDispatcher", "receiveRequest", "Thread: "+ Thread.currentThread().getName() +"; EXCEPTION: "+ e.toString());}
			
		this.Request = new TRequest(Request);
		this.SocketWaitingResult = SocketWaitingResult;
		notReady = true;
		notifyAll();		
	}
	        
        /** Runs iAccessDispatcher in a new thread.*/
	public void run() {
//		Start JBOSS Application Server in a new thread
//		If JBOSS receives a request it calls the receiveRequest method of the iAccessDispatcher
		JBOSS JTCPServer = new JBOSS(this);
		
		try {			
			synchronized (this) {
				JTCPServer.start();
				stopWorking = false;
				while (true){		
					try{		
						notReady = false;
						wait();
						notifyAll();
					} catch (Exception e){
						iAccessDispatcher.logger.logp(Level.SEVERE, "iAccessDispatcher", "run", "Thread: "+ Thread.currentThread().getName() +"; EXCEPTION: "+ e.toString());
					}
					if (stopWorking) return;
					InteractionCounter ++;
					if (Request.IsService) {
                                            if (TSessionInfo.IsServerAuthorizationMode) {
                                                System.out.println("IACCESS: Server authz mode: Adding the request to a pending list.");
                                                addPendingRequest(this.Request, this.SocketWaitingResult);
                                            } else 
                                                startNewSessionServer(this.Request, this.SocketWaitingResult);
                                        } else 
						joinExistSession();
				}
			}		
		} catch (Exception e) {
			iAccessDispatcher.logger.logp(Level.SEVERE, "iAccessDispatcher", "run", "Thread: "+ Thread.currentThread().getName() +"; EXCEPTION: "+ e.toString());
		} finally {			
			JTCPServer.Stop();
			Finalize();
		}		
	}
        
        /** Sends stopWorking massage to iAccessDispatcher.*/
	public synchronized void stopServer(){
		stopWorking = true;
		notifyAll();		
	}
        
        /** Finalizes all currently active sessions and finally shuts down the iAccessDispatcher.*/
	private void Finalize(){
		try {
//			Finalize all sessions and threads. Deem that all sessions go one after another. 
			long iAccessDispatcherReleaseTime = iAccessDispatcher.SessionInfoTable.size() * TSessionInfo.SessionTimeLimit + System.currentTimeMillis();
		
			while ( iAccessDispatcherReleaseTime > System.currentTimeMillis()){			
				while (iAccessDispatcher.SessionInfoTable.elements().hasMoreElements() ){
					ThreadGroup CurrThreadGroup = ((TSessionInfo) iAccessDispatcher.SessionInfoTable.elements().nextElement()).TNThreadGroup;
					Thread[] CurrSessionThreads = new Thread[CurrThreadGroup.activeCount()];
					CurrThreadGroup.enumerate(CurrSessionThreads);
					for (int i = 0; i < CurrSessionThreads.length; i++)
						try {
							if ((CurrSessionThreads[i] != null) && (CurrSessionThreads[i].isAlive()) )
								CurrSessionThreads[i].join();
						} catch (Exception e) {
							iAccessDispatcher.logger.logp(Level.SEVERE, "iAccessDispatcher", "Finalize", "Thread: "+ Thread.currentThread().getName() +"; EXCEPTION: "+ e.toString());
						}					
				}
				if (SessionInfoTable.size() == 0) 
                                    break;
                                else
                                    Thread.currentThread().sleep(UnexpectedErrorSleepTime);
			}			
		} catch (Exception e) {
			iAccessDispatcher.logger.logp(Level.SEVERE, "iAccessDispatcher", "Finalize", "Thread: "+ Thread.currentThread().getName() +"; EXCEPTION: "+ e.toString());
		} finally {
			SessionInfoTable.clear();
			iAccessDispatcher.logger.logp(Level.INFO, "iAccessDispatcher", "Finalize", "iAccessDispatcher is destroyed");
			LoggerHandler.close();			
		}
	}

        

        
        
        /* START iAccess Server Mode Integration Part.
         Little internal convention:
           '-' means an access decision is not ready for a current request  
           '1' means an access deciison is ready and granted
           '2' means an access decision is ready and denied
         */
         /**
          * Stores information on received service requests by iAccess agents. This data is used later
          * by the function getAccessDecisionInServerMode() to complete pending negotiations.
          */
        public static Hashtable pendingRequestList = new Hashtable();
        
        /**
         * Initiates a server side negotiations of pending service requests received by client iAccess agents.
         * The main function by an internal/local PEP server to ask for an access decision.
         * The function returns a boolean value if the service request is granted or not.
         * If you need a SAML decision statement then you should add inside a call to a SAML
         * AuthzDecisionStatement.
         * <p>This method is supposed to be invoked by a PEP internal to the iAccess agent platform.
         * To use the method you should turn on the respective field in the main configuration file (IACCESS_ROOT/config.xml).
         * @param serverRequest a service request generated by a PEP. You should specify only SessionID and IsService=true fields. The other parameters are already stored by yhte iAccess agent, if a such a request has arrived.
         * @return grant/deny of a negotiation process.
         */
        public boolean getAccessDecisionInServerMode(TRequest serverRequest) {
            System.out.println("IACCESS: PEP: Received request from a PEP server with a SessionID: "+serverRequest.SessionID);
            boolean result = false; 
            if ( isPending(serverRequest) && serverRequest.IsService) {
                Object[] data = (Object[]) pendingRequestList.get( serverRequest.SessionID + "-" );
                System.out.println("IACCESS: PEP: Matched pending request and starts a negotiation.");
               //iAccess starts negotiations and awaiting for the access decision
               //Starts a new session as server.
               startNewSessionServer((TRequest)data[0], (SSLSocket)data[1]);
                //await for negotiation resut  
                try {
                   while ( isPending(serverRequest) ) 
                    Thread.currentThread().sleep(50);
                } catch (Exception e){
                    System.err.println("IACCESS: EXCEPTION: getAccessDecisionInServerMode: " + e.toString());    
                }
               
               if ( pendingRequestList.containsKey( serverRequest.SessionID + "0")) {
                    result = false;
                    pendingRequestList.remove(serverRequest.SessionID + "0");
               }
               
               if ( pendingRequestList.containsKey( serverRequest.SessionID + "1")) {
                    result = true;
                    pendingRequestList.remove(serverRequest.SessionID + "1");
               }
            }
            else
            {
                System.out.println("IACCESS: WARNING: Did not match a pending request for a SessionID: "+serverRequest.SessionID);
            }
            return result;
        }
        



        /** The main function by an internal/local PEP server to initiate an access decision (negotiation) process.
         * The function returns a set of active credentials negotiated upon with an opponent.
         * The only important fields in the input to this function are the SessionID and IsService=true fields.
         * To use the method you should turn on the respective field in the main configuration file (IACCESS_ROOT/config.xml).
         * <p>This function has been added in response to iAccess integration within a project EU-FP7-OKKAM.
         * Importantly here is not a decision of grant/deny but a set of credentials agreed with an opponent.
         * Thus, given an OKKAM query (database) result one can select all objects of the query result that opponent's
         * active credentials satisfy (what abjects are visible to a client of the query), regardless of the final negotiation decision.
         * @param serverRequest a service request generated by a PEP. You should specify only SessionID and IsService=true fields. The other parameters are already stored by yhte iAccess agent, if a such a request has arrived.
         * @return the set of presented credentias by an opponent out of the negotiation for the service request. (Vector of strings, logic format credentials)
         */
        public Vector getAccessDecisionInOKKAMMode(TRequest serverRequest) {
            System.out.println("IACCESS: OKKAM: Received request from SAC with a SessionID: "+serverRequest.SessionID);
            Vector result = null;
            if ( isPending(serverRequest) && serverRequest.IsService) {
                Object[] data = (Object[]) pendingRequestList.get( serverRequest.SessionID + "-" );
                System.out.println("IACCESS: OKKAM: Matched pending request and starts a negotiation.");
               //iAccess starts negotiations and awaiting for the access decision
               //Starts a new session as server and awaits for the credentials out of the negotiation
               startNewSessionServer((TRequest)data[0], (SSLSocket)data[1]);
                try { 
                   while ( isPending(serverRequest) )
                    Thread.currentThread().sleep(50);
                } catch (Exception e){
                    System.err.println("IACCESS: EXCEPTION: getAccessDecisionInOKKAMMode: " + e.toString());
                }

               if ( pendingRequestList.containsKey( serverRequest.SessionID + "0")) {
                   //take credentials of denial completed negotiation
                   result = (Vector) pendingRequestList.get(serverRequest.SessionID + "0");
                   pendingRequestList.remove(serverRequest.SessionID + "0");
               }

               if ( pendingRequestList.containsKey( serverRequest.SessionID + "1")) {
                   //take credentials of successfully completed negotiation
                   result = (Vector) pendingRequestList.get(serverRequest.SessionID + "1");
                   pendingRequestList.remove(serverRequest.SessionID + "1");
               }
            }
            else
            {
                System.out.println("IACCESS: WARNING: Did not match a pending request for a SessionID: "+serverRequest.SessionID);
            }
            return result;
        }



        /**
         * Compares requests from a PEP and an iAccess client agent.
         */
        private boolean isPending(TRequest Request) {
            
            return pendingRequestList.containsKey( Request.SessionID+"-");
        }

        /**
         * Adds to a hashtable an incoming request from a remote iAccess agent.
         * It adds also the object of the active socket connection, which is retrieved later on by
         * PEP invocations.
         */
        private void addPendingRequest(TRequest Request, SSLSocket SocketWaitingResult) {
            //New Request from an iAccess client agent is received.
            TRequest Temp = new TRequest(Request);
            Object[] data = new Object[2];
            data[0] = Temp;
            data[1] = SocketWaitingResult;
            pendingRequestList.put( Temp.SessionID+"-", data); 
        }
        
        /* END iAccess Server Mode */
        
}