/*
 * Copyright 2006, 2008 Hristo Koshutanski, Aliaksandr Lazouski, Associazione CREATE-NET, and Universita' degli Studi di Trento.  
 * 
 * This file is part of iAccess authorization service software (called iAccess).
 * iAccess is free software: you can redistribute it and/or modify 
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * iAccess is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with iAccess.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.iaccess.TNProtocol;

import java.util.Vector;
import java.util.Calendar;
import java.util.logging.Level;
import javax.net.ssl.SSLSocket;

import org.iaccess.LogicalLayer.BasicAccessControlAlgorithm;
import org.iaccess.LogicalLayer.ResultOfOperation;
import org.iaccess.CredentialManager.*;

/**
 * Manages a single interaction in a negotiation within a given session.
 * If a request is already being managed by another object, iAccessNegotiation joins to this object and awaits for the result (suspended thread on already requested credential), 
 * else it calls iAccess logical layer to get the result. 
 * In this case iAccessNegotiation can generate some counter-requests and consequently start some new interactions in new threads. 
 * iAccessNegotiation calls TNProtocol.JBOSS server methods to send requests.
 * The object of iAccessNegotiation will finalize its work when the Result of the request (grant or deny) is taken. 
 * This class extends the abstract class java.lang.Thread, threfore always starts in a new thread.
 * <p> This class encapsulates the negotiation protocol functionality as presented in the JNSM paper.
 * The class implements AskCredentials() function with the stepwise disclosure.
 * <p>More details are given to each of the functions in the class.
 * @see org.iaccess.TNProtocol.iAccessDispatcher
 * @see org.iaccess.LogicalLayer.BasicAccessControlAlgorithm
 */
public class iAccessNegotiation extends Thread {

	/** Response value for grant. */
	public static final int grant = 1;
	/** Response value for deny. */
	public static final int deny = 0;
	/** Response value for not ready. */
	public static final int notReady = -1;
    /** Keeps the result of a current negotiation. */
    private int Result = notReady;
        /** Is false if such request is not being managed now*/
	private boolean IsSuspended = false;        	
        /**Specifies the current processing request*/
	private TRequest Request;
        /**Defines the reference to the SessionInfo of current processing request*/
	private TSessionInfo SessionInfo;
        /** Specifies the addressee of the request.
        * TCP/IP sockets are used. This variable should be replaced in the next version by more complete value */
	private SSLSocket SocketWaitingResult;
        /** Consists the links for all iAccessNegotiation entities joined this object for the result*/
	private Vector SuspendedThreads;
        /** Global counter of all existing interactions. It is unique for earch interaction. 
         * We use it to distinguish temporary files produced by iAccess logical layer within one session*/
	private long InteractionCounter = 0;
	/** Specifies an instance of iAccess Logical Layer*/
	private BasicAccessControlAlgorithm ObjBasicAccessControlAlgorithm = new BasicAccessControlAlgorithm();

    /** Returns the global counter of all existing interactions. It is unique for earch interaction.
     * We use it to distinguish temporary files produced by iAccess logical layer within one session*/
    public long getCounter() {
		return InteractionCounter;
	}

    /** Local class implementing the possibility to send counter-requests. It is used from inside the AskCredentials() function.
     * All missing credentials are requested in new threads using this class.
     */
	private class TRequestThread implements Runnable{

		private TRequest Request; 
		private int Result = deny;
		/** Runs a credential request in a new thread. */
		public TRequestThread(TRequest Request, ThreadGroup threadgroup){
			this.Request = Request;
			Thread thread = new Thread(threadgroup, this, "CR: "+ Request.SessionID + " "+ Request.ServiceRequest + InteractionCounter);
			thread.start();
		}
		/** Runs a credential request in a current thread. It is used for the last credential in a missing set. */
		public TRequestThread(TRequest Request){
			this.Request = Request;
			run();			
		}
		/** Send counter-request to a server. */
		public void run(){
			try {
                                boolean flIsAnyValueRequest = false;
//				Send Request to JBOSS and update SessionInfo				
                                String newResult = null;
				try {
                                newResult = JBOSS.sendRequestSSL(Request, SocketWaitingResult.getInetAddress());					
				} catch (Exception e){
					Result = deny; newResult = null;
					iAccessDispatcher.logger.logp(Level.WARNING, "TRequestThread", "run", "Thread: "+ Thread.currentThread().getName() +"; EXCEPTION: "+ e.toString());				
				}
			       // PROCESSING A RECEIVED RESPONSE 
				if ( newResult != null )//if grant	
                                {
                                        //update with anyvalueterm instsntiation response 
                                        if(newResult.length() > "grant".length())
                                        {
                                            ((TSessionInfo)iAccessDispatcher.SessionInfoTable.get(Request.SessionID)).UpdateCP(newResult.substring("grant".length()));
                                            flIsAnyValueRequest = true;
                                        }
                                        //update with what was the request for a credential
                                        ((TSessionInfo)iAccessDispatcher.SessionInfoTable.get(Request.SessionID)).UpdateCP(Request.ServiceRequest);
                                        Result = grant; 
                                } else { 
                                        //update also with anyvalue response *by HRISTO*
                                        //Actually, no neeed of the above as the declined cred is the avyvalue of (if used)
					((TSessionInfo)iAccessDispatcher.SessionInfoTable.get(Request.SessionID)).UpdateCN(Request.ServiceRequest);
                                        Result = deny;
                                }
				if (iAccessDispatcher.OutPutStream != null){
					String TempString = "notReady";
					if (Result == deny) TempString = "denied";
					if (Result == grant) TempString = "granted";
                                        if(flIsAnyValueRequest)
                                            iAccessDispatcher.OutPutStream.write( ("<<<"+ newResult.substring("grant".length()) + "!" + "incoming credential response: "+TempString + "}" + Request.SessionID).getBytes());
                                        else
                                            iAccessDispatcher.OutPutStream.write( ("<<<"+ Request.ServiceRequest + "!" + "incoming credential response: "+TempString + "}" + Request.SessionID).getBytes());
                                            
				}

			} catch (Exception e) {
				iAccessDispatcher.logger.logp(Level.WARNING, "TRequestThread", "run", "Thread: "+ Thread.currentThread().getName() +"; EXCEPTION: "+ e.toString());
			}
		}
	}
/**
 * Constructor of the class. Initializing main data for the negotiation protocol.
 */
	public iAccessNegotiation(TRequest Request, SSLSocket SocketWaitingResult){
		super( ((TSessionInfo) iAccessDispatcher.SessionInfoTable.get(Request.SessionID)).TNThreadGroup, 
				"TN: "+Request.SessionID+" " + Request.ServiceRequest+" "+ iAccessDispatcher.InteractionCounter);
                //Unique for each interaction
		InteractionCounter = iAccessDispatcher.InteractionCounter;
		this.SocketWaitingResult = SocketWaitingResult; 
		this.Request = Request;
		this.SuspendedThreads = new Vector();
		this.SessionInfo = ( (TSessionInfo)iAccessDispatcher.SessionInfoTable.get(Request.SessionID));             
	}
        /** Starts negotiations on the received request. It runs the iAccessNegotiation in a new thread.
         * iAccessNegotiation checks if a request is to be suspended, and, if not, it runs the negotiation protocol
         * (calls runTNProtocol() function).
         */
	public void runNegotiations(){		
		start();
	}

    public void run(){	//runNegotiations()
 		try {
			if(!Request.IsService){
//				If such Request for the Credential is already in a negotiation ...
				SessionInfo.IncreaseBlockCnegCounter();
				iAccessNegotiation threadInNeg = JoinToThread(Request.ServiceRequest);
				if (threadInNeg != null){						
					IsSuspended = true;
//					Suspend and await for result
					if (iAccessDispatcher.OutPutStream != null)
						iAccessDispatcher.OutPutStream.write( ("<<<"+Request.ServiceRequest+"!incoming credential request: suspended" + "}" + Request.SessionID).getBytes());

					Result = threadInNeg.waitForResult();
					if (Result == notReady) Result = deny;
				} else { 					
					IsSuspended = false;
					SessionInfo.DecreaseBlockCnegCounter();
				}
			}			
                        //Runs trust negotiation protocol
			if (Result == notReady) {
                                String tmpstr = null;
                                if(Request.IsService) tmpstr = "service"; else tmpstr = "credential";
				if (iAccessDispatcher.OutPutStream != null)
					iAccessDispatcher.OutPutStream.write( ("<<<"+Request.ServiceRequest+"!incoming "+tmpstr+" request: processed" + "}" + Request.SessionID).getBytes());
				Result = runTNProtocol();
			}		
			
		} catch (Exception e) {
			iAccessDispatcher.logger.logp(Level.WARNING, "iAccessNegotiations", "run/runNegotiations()", "Thread: "+ Thread.currentThread().getName() +"; EXCEPTION: "+ e.toString());
			Result = deny;
		} finally {
			Finalize();
		}
	}
        /** Runs the trust negotiation protocol that calls the core interactive access control
         * function of the logical layer. 
         * The protocol steps are in accordance with the negotiation scheme of JNSM paper.
         * @see org.iaccess.LogicalLayer.BasicAccessControlAlgorithm
         */
	public int runTNProtocol() {   
                //BY THE TIME BEING THIS LINE ADDS AN EMPTY SET
                //IN NEXT RELEASE THIS LINE WOULD ADD INITIAL CHAIN OF CREDENTIALS.
                //WE FIXED Credentials[0] to return CredAnyValueTerm for synchronization
                //WE SHOULD ADD FROM INDEX 1
                //Uncomment the line below when necessary for extensions.
		//SessionInfo.UpdateCP(Request.Credentials[XX]);
		int Result = notReady;
		Vector CM = new Vector();

		for (;;){			
                        //Is Session expired?
			if (System.currentTimeMillis() >= SessionInfo.GetValidNotAfter() ) 				
				return deny;				
			try {
                                if(TSessionInfo.MultiSessionSharedCreds)
                                {
                                    long ln1 = SessionInfo.GetValidNotAfter();
                                    long ln2 = SessionInfo.negStartTime;
                                    if(ln2+TSessionInfo.SessionTimeLimit < ln1)
                                       SessionInfo.UpdateCP(UserProfile.getCredentials(Request.Identity, ln1));
                                    else
                                       SessionInfo.UpdateCP(UserProfile.getCredentials(Request.Identity, ln1+TSessionInfo.DynamicSessionTimeExtension));
                                }
                                else 
                                    //We want to update the identity token of an opponent from the SSL channel. The token is alreadt cheched for validity against current session in TNProtocol.JBOSS Handshakelistener event.
                                    //If already updated before, it will not be added again.  
                                    SessionInfo.UpdateCP(UserProfile.getEntityLogicIDCredOfSSL(Request.Identity));
                                
                                //take access decision
				Result = iAccessDecision(SessionInfo.GetCP(), SessionInfo.GetCN(), CM);
				switch (Result) 
                {
					case grant: case deny:
					     return Result; 
					case notReady:
					     AskCredentials(CM);
                } 
                                // if CM \not \subseteq CP U CN
                                // We do the check to ensure that if AskCredentials() function exits in a unpredicted way,
                                // or some missing credentials are neither presented nor declined (for whatever reason)
                                // then we mark then declined.
                                if(!IsSubsetOrEqualOfTwoJointSets(CM,SessionInfo.GetCP(),SessionInfo.GetCN()))
                                {   Vector cCP = SessionInfo.GetCP();
                                    for(int i=0; i<CM.size();i++)
                                    {
                                        if(!cCP.contains(CM.get(i).toString()))
                                            SessionInfo.UpdateCN(CM.get(i).toString());
                                    }
                                }
			} catch (Exception e){
				iAccessDispatcher.logger.logp(Level.WARNING, "iAccessNegotiations", "runTNProtocol", "Thread: "+ Thread.currentThread().getName() +"; EXCEPTION: "+ e.toString());
				return deny;
			}			
		}
	}
	
        /**
         * Internal use for the negotiation protocol (runTNProtocol()).
         */
        private boolean IsSubsetOrEqualOfTwoJointSets(Vector TargetSet, Vector DestSet1, Vector DestSet2)
        {
            //There are no repeating elements in either of the three sets by assumption
            //The two destination sets are disjoint by assumption
            boolean flag4success = false;
            int successfulhits=0;

            for(int i=0; i<TargetSet.size(); i++)
            {
                flag4success = false;
                for(int j=0;j<DestSet1.size();j++)
                {
                    if(TargetSet.get(i).toString().equals(DestSet1.get(j).toString()))
                    {
                        flag4success = true;
                        break;
                    }
                }
                if(flag4success)
                {
                    successfulhits++;
                    continue;
                }
                
                for(int j=0;j<DestSet2.size();j++)
                {
                    if(TargetSet.get(i).toString().equals(DestSet2.get(j).toString()))
                    {
                        flag4success = true;
                        break;
                    }
                }
                if(flag4success)
                    successfulhits++;
            }

            if(TargetSet.size() == successfulhits)
                return true;
            else
                return false;
        }
        
    /**
     * Transforms a set of missing credentials to set of requests to an opponent. It uses the stepwise
     * disclosure functionality, by calling doStepwiseDisclosure() of the logical layer.
     * @param MissCreds a set of missing credentials (Vector of strings).
     * @see org.iaccess.LogicalLayer.BasicAccessControlAlgorithm
     * @see org.iaccess.TNProtocol.iAccessNegotiation.TRequestThread
     */
	private void AskCredentials(Vector MissCreds) 
        {
            try {
                do{
                    //Begin Stepwise Loop
                    //Update user profile if very recent (multi) session negotiations.
                    if(TSessionInfo.MultiSessionSharedCreds)
                    {
                        long ln1 = SessionInfo.GetValidNotAfter();
                        long ln2 = SessionInfo.negStartTime;
                        if(ln2+TSessionInfo.SessionTimeLimit < ln1)
                           SessionInfo.UpdateCP(UserProfile.getCredentials(Request.Identity, ln1));
                        else
                           SessionInfo.UpdateCP(UserProfile.getCredentials(Request.Identity, ln1+TSessionInfo.DynamicSessionTimeExtension));
                    }
                    else 
                        //We update the identity token of an opponent from the SSL channel. The token is alreadt cheched for validity against current session in TNProtocol.JBOSS Handshakelistener event.
                        //If already updated before, it will not be added again.  
                        SessionInfo.UpdateCP(UserProfile.getEntityLogicIDCredOfSSL(Request.Identity));
                    
                    if(IsSubsetOrEqualOfTwoJointSets(MissCreds,SessionInfo.GetCP(),SessionInfo.GetCN()))
                        return;
                    if (!ObjBasicAccessControlAlgorithm.SetUp(Request, SessionInfo.ServiceRequest, SessionInfo.GetCP(), SessionInfo.GetCN(), SessionInfo.CredAnyValueTerm, SessionInfo.negStartTime).isSuccessful())        	
                        return;    
                    ResultOfOperation dostwsd = ObjBasicAccessControlAlgorithm.doStepwiseDisclosure(MissCreds);
                    if (!dostwsd.isSuccessful())
                        return;
                    if (dostwsd.isSuccessful() && (dostwsd.getOperationResult()==null))
                        return;
                    //take the one-step missing credentials
                    Vector CM = (Vector) dostwsd.getOperationResult();

                    //For one credential and/or last credential we use current thread
                    //TNThreadGroup keeps information about all generated threads                    
                    ThreadGroup TNThreadGroup = new ThreadGroup("AskCredentials");		                        

                        for(int i = CM.size()-1; i > 0; i--) 
                        {
                            if( TSessionInfo.MultiSessionSharedCreds &&
                                (CM.get(i).toString().indexOf(TSessionInfo.CredAnyValueTerm)==-1))
                            { 
                                long ln1 = SessionInfo.GetValidNotAfter();
                                long ln2 = SessionInfo.negStartTime;
                                if(ln2+TSessionInfo.SessionTimeLimit < ln1)
                                {
                                   if(UserProfile.containts((String) CM.get(i),ln1))
                                   {
                                       SessionInfo.UpdateCP((String) CM.get(i));
                                       continue;
                                   }
                                }
                                else
                                {
                                   if(UserProfile.containts((String) CM.get(i),ln1+TSessionInfo.DynamicSessionTimeExtension))
                                   {
                                       SessionInfo.UpdateCP((String) CM.get(i));
                                       continue;
                                   }
                                }
                            }
                            if (iAccessDispatcher.OutPutStream != null)
                                iAccessDispatcher.OutPutStream.write( (">>>"+((String)CM.get(i))+"!outgoing credential request: submitted" + "}" + Request.SessionID).getBytes());
                            new TRequestThread(new TRequest(Request.SessionID,false, (String) CM.get(i), CredentialsConfig.principalName, TSessionInfo.CredAnyValueTerm, TSessionInfo.TCPportIn),TNThreadGroup);
			}
		
                        //Request the first credential in the current thread
                        boolean fl = false;
                        if( TSessionInfo.MultiSessionSharedCreds && 
                            (CM.get(0).toString().indexOf(SessionInfo.CredAnyValueTerm)==-1))
                        {

                                long ln1 = SessionInfo.GetValidNotAfter();
                                long ln2 = SessionInfo.negStartTime;
                                if(ln2+TSessionInfo.SessionTimeLimit < ln1)
                                {
                                   if(UserProfile.containts((String) CM.get(0),ln1))
                                   {
                                       SessionInfo.UpdateCP((String) CM.get(0));
                                       fl = true;
                                   }
                                }
                                else
                                {
                                   if(UserProfile.containts((String) CM.get(0),ln1+TSessionInfo.DynamicSessionTimeExtension))
                                   {
                                       SessionInfo.UpdateCP((String) CM.get(0));
                                       fl = true;
                                   }
                                }
                        
                        }
                        if(!fl)
                        {
                            if (iAccessDispatcher.OutPutStream != null)
                            	iAccessDispatcher.OutPutStream.write( (">>>"+ ((String)CM.get(0))+"!outgoing credential request: submitted" + "}" + Request.SessionID).getBytes());
                            new TRequestThread(new TRequest(Request.SessionID,false, (String) CM.get(0), CredentialsConfig.principalName, TSessionInfo.CredAnyValueTerm, TSessionInfo.TCPportIn));
                        }
                    
                        //Awaits for all generated thread untill be completed
			Thread RequestThreads[] = new Thread[TNThreadGroup.activeCount()];
			TNThreadGroup.enumerate(RequestThreads);
				
			for (int i = 0; i< RequestThreads.length; i++) {
                            try {
                                if (RequestThreads[i].isAlive())
                                    RequestThreads[i].join();
                            } catch (Exception e) {
                                iAccessDispatcher.logger.logp(Level.WARNING, "iAccessNegotiations", "AskCredentials", "Thread: "+ Thread.currentThread().getName() +"; EXCEPTION: "+ e.toString());
                            }		
			}
            
                //End Stepwise loop
                // condition for loop breaking is CM==null which is 
                // checked at the beginning of the loop.
                // doStepwiseDisclosure() guarantees that
                // it will return null at some point of iteractions.
                }while(true);
            }
            catch (Exception e) 
            {
                iAccessDispatcher.logger.logp(Level.WARNING, "iAccessNegotiations", "AskCredentials", "Thread: "+ Thread.currentThread().getName() +"; EXCEPTION: "+ e.toString());
            }
	}
        
        
        
	/** Searches for an existing objects which have been processing the same request. Null is returned if nothing is found.
     * If nothing is found we start a new negotiations process denoted by changing the state of object (name).
     * Now all new objects with the same request will be suspended on this object.
     */
	private iAccessNegotiation JoinToThread( String Credential){
		synchronized (SessionInfo.TNThreadGroup) {
			try {
//                              Links for all existing objects within specified session
				Thread CurrSessionThreads[] = new Thread[SessionInfo.TNThreadGroup.activeCount()];
				SessionInfo.TNThreadGroup.enumerate(CurrSessionThreads);
				for (int i = 0; i < CurrSessionThreads.length; i++) {
					if ( (CurrSessionThreads[i] != null) && (CurrSessionThreads[i].isAlive()) && (CurrSessionThreads[i].getName().startsWith("*TN: "+ Request.SessionID +" "+Credential)) ) 
						return (iAccessNegotiation) CurrSessionThreads[i];			
				}
//                              If nothing has found we start a new negotiations process denoted by changing the state of object (name).
//                              Now all new objects with the same request will suspend on this object                                
				Thread.currentThread().setName("*" + Thread.currentThread().getName());
				return null;
			} catch (Exception e){
				iAccessDispatcher.logger.logp(Level.WARNING, "iAccessNegotiations", "InNegotiation", "Thread: "+ Thread.currentThread().getName() +"; EXCEPTION: "+ e.toString());
				return null;
			}
		}
	}
	/** Method is called by other threads with a same request as the current object (negotiation object).
     * All these threads await for an access decision (a result) by the current object.
     * In case of a dead loop (when session time is expired) we prolong the current session with the DynamicSessionTimeExtension interval
     */
	public int waitForResult(){
		long ValidNotAfter = SessionInfo.GetValidNotAfter();
		try {			 
			SuspendedThreads.add(Thread.currentThread());
			synchronized (this) {
				SessionInfo.DecreaseBlockCnegCounter();
                                wait( ValidNotAfter - System.currentTimeMillis() );
			}			
		} catch (Exception e ) {
			iAccessDispatcher.logger.logp(Level.WARNING, "iAccessNegotiations", "waitForResult", "Thread: "+ Thread.currentThread().getName() +"; EXCEPTION: "+ e.toString());
		}
		//In case of dead loop (when session time is expired) we prolong session for DynamicSessionTimeExtension interval
		if (Result == notReady) {
			if (ValidNotAfter == SessionInfo.GetValidNotAfter())
				SessionInfo.SetValidNotAfter(ValidNotAfter + TSessionInfo.DynamicSessionTimeExtension );				
			Result = deny;
		}
		return Result;
	}
	/** Notifies all suspended threads that a decision for the current request has been taken.
     * It awaits for some time required to correct finalization of all alien threads.
     */
	private void NotifySuspendedThreads() {
		synchronized (this) {
			notifyAll();	
		}
				
		for (int i = 0; i < SuspendedThreads.size(); i++) {
			if (((Thread) SuspendedThreads.get(i)).isAlive())
				try {
					((Thread) SuspendedThreads.get(i)).join();					
				} catch (Exception e) {
					iAccessDispatcher.logger.logp(Level.WARNING, "iAccessNegotiations", "NotifySuspendedThreads", "Thread: "+ Thread.currentThread().getName() +"; EXCEPTION: "+ e.toString());
				}
		}
	}

    /** Finalizes a current negotiation protocol process.*/
	private void Finalize(){
		if (Result == notReady)	Result = deny;
		try {
//			Remove Cneg taking into account risk of error result (Step 16 - Step 3 of figure in JNSM paper)
			if ((!IsSuspended) && (!Request.IsService)) 
				SessionInfo.RemoveCneg();
			
//			Notify all awaiting for r'result threads
			NotifySuspendedThreads();
			JBOSS.sendResponseSSL(Request, Result, SocketWaitingResult);
	
//			if Session is finalizing remove SessionInfo 
                        String negTime = null;
			if (Request.IsService) {
                            /* Begin iAccess integration with a Server Authz Mode */
                            if (TSessionInfo.IsServerAuthorizationMode) {
                                iAccessDispatcher.pendingRequestList.put( Request.SessionID + String.valueOf(Result), ( (TSessionInfo) iAccessDispatcher.SessionInfoTable.get(Request.SessionID)).GetCP());
                                iAccessDispatcher.pendingRequestList.remove(Request.SessionID + "-");
                            }
                            /* End iAccess integration with a Server Authz Mode */
                            
                            
//				Await for all other threads in session
				Thread CurrSessionThreads[] = new Thread[SessionInfo.TNThreadGroup.activeCount()];
				SessionInfo.TNThreadGroup.enumerate(CurrSessionThreads);
				for (int i = 0; i < CurrSessionThreads.length; i++) { 
					if ( (CurrSessionThreads[i] != null) && (!CurrSessionThreads[i].equals(Thread.currentThread())) && CurrSessionThreads[i].isAlive())
						CurrSessionThreads[i].join();
				}
                negTime = Long.toString(Calendar.getInstance().getTimeInMillis() - ((TSessionInfo)iAccessDispatcher.SessionInfoTable.get(Request.SessionID)).negStartTime );
                System.out.println("IACCESS: Negotiation Time (ms):" + negTime);
                iAccessDispatcher.SessionInfoTable.remove(Request.SessionID);
                                
			}
			
			if (iAccessDispatcher.OutPutStream != null) {
				String TempString = "notReady";
				if (Result == deny) TempString = "denied";
				if (Result == grant) TempString = "granted";
                                if (Request.IsService)
                                {
				    iAccessDispatcher.OutPutStream.write( (">>>"+Request.ServiceRequest + "!" + "outgoing service response: "+TempString + " (" + negTime + " ms)" + "}" + Request.SessionID).getBytes());
                                }
                                else 
                                {
                                    iAccessDispatcher.OutPutStream.write( (">>>"+Request.ServiceRequest + "!" + "outgoing credential response: "+TempString + "}" + Request.SessionID).getBytes());
                                }
			}
		} catch (Exception e) {
			iAccessDispatcher.logger.logp(Level.SEVERE, "iAccessNegotiations", "Finalize", "Thread: "+ Thread.currentThread().getName() +"; EXCEPTION: "+ e.toString());
		}
	}
        
        /** Calls functions of logical layer to get the result of a request.
         * It conforms to the negotiation scheme presented in JNSM paper.
         * @param CP a current set of presented credentials.
         * @param CN a current set of declined credentials.
         * @param CM a set of missing credentials from the previous interaction.
         * @return grant/deny/notReady (a missing set of credentials)
         */
	private int iAccessDecision(Vector CP, Vector CN, Vector CM)throws Exception {         
		int Result = iAccessNegotiation.deny;

                if (!BasicAccessControlAlgorithm.IsInitialized) 
                {
	        String DLVWrapperConfigurationFile = TSessionInfo.DLVWrapperConfigurationFile;
		boolean MinimalityCriterion;
		if ( TSessionInfo.MinimalityCriterion.equals("ON") ) {
			MinimalityCriterion = true;
		} else if( TSessionInfo.MinimalityCriterion.equals("OFF") ) {
			MinimalityCriterion = false;
		}else {
			throw new Exception("The specified value of MinimalityCriterion does not suite!");
		}
                    ObjBasicAccessControlAlgorithm.LoadConfiguration(DLVWrapperConfigurationFile, MinimalityCriterion, TSessionInfo.ConfigRootDir, TSessionInfo.RootDir, TSessionInfo.OSType);	        
		}     
                
                if (!ObjBasicAccessControlAlgorithm.SetUp(Request, SessionInfo.ServiceRequest, CP, CN, SessionInfo.CredAnyValueTerm, SessionInfo.negStartTime).isSuccessful())        	
                    return Result;    
                
                ResultOfOperation resAccessDecision = ObjBasicAccessControlAlgorithm.computeAccessDecision();                
                
                Vector tmpVecCM= new Vector();
                if(resAccessDecision.isSuccessful()) {
        	try {
        		Integer explicitResult = (Integer) resAccessDecision.getOperationResult();
        		if(explicitResult.compareTo(new Integer(1)) == 0)
        			Result = iAccessNegotiation.grant;
        		else
        			Result = iAccessNegotiation.deny;
        	} catch(ClassCastException e) {
        		tmpVecCM = (Vector) resAccessDecision.getOperationResult();
        		CM.clear();
        		for(int j = 0; j < tmpVecCM.size();j++) {
        			CM.add(tmpVecCM.get(j));	
        		}
        		Result = iAccessNegotiation.notReady;
        	}
        }
        return Result;        
	}	
}